# © Copyright Databand.ai, an IBM Company 2022
from distutils.version import LooseVersion

import airflow

from airflow.configuration import conf
from airflow.exceptions import AirflowConfigException


AIRFLOW_VERSION_2 = LooseVersion(airflow.version.version) >= LooseVersion("2.0.0")
AIRFLOW_VERSION_1 = LooseVersion(airflow.version.version) < LooseVersion("2.0.0")

AIRFLOW_VERSION_BEFORE_2_2 = LooseVersion(airflow.version.version) < LooseVersion(
    "2.2.0"
)
AIRFLOW_VERSION_AFTER_2_2 = LooseVersion(airflow.version.version) >= LooseVersion(
    "2.3.0"
)

AIRFLOW_ABOVE_13 = LooseVersion(airflow.version.version) > LooseVersion("1.10.13")
AIRFLOW_ABOVE_10 = LooseVersion(airflow.version.version) > LooseVersion("1.10.10")
AIRFLOW_ABOVE_9 = LooseVersion(airflow.version.version) > LooseVersion("1.10.9")
AIRFLOW_ABOVE_6 = LooseVersion(airflow.version.version) > LooseVersion("1.10.6")
AIRFLOW_BELOW_10 = LooseVersion(airflow.version.version) < LooseVersion("1.10.10")
if AIRFLOW_VERSION_1:
    from airflow.hooks.base_hook import BaseHook

elif AIRFLOW_VERSION_2:
    from airflow.hooks.base import BaseHook


def get_task_log_reader():
    try:
        return conf.get("core", "task_log_reader")
    except AirflowConfigException:
        return conf.get("logging", "task_log_reader")


__all__ = ["BaseHook"]



# © Copyright Databand.ai, an IBM Company 2022

import os

from dbnd._core.configuration.environ_config import ENV_DBND_HOME, ENV_DBND_SYSTEM
from dbnd._core.log import dbnd_log_init_msg


def _initialize_airflow_home():
    ENV_AIRFLOW_HOME = "AIRFLOW_HOME"

    if ENV_AIRFLOW_HOME in os.environ:
        # user settings - we do nothing
        dbnd_log_init_msg(
            "Found user defined AIRFLOW_HOME at %s" % os.environ[ENV_AIRFLOW_HOME]
        )
        return

    for dbnd_airflow_home in [
        os.path.join(os.environ[ENV_DBND_SYSTEM], "airflow"),
        os.path.join(os.environ[ENV_DBND_HOME], ".airflow"),
    ]:
        if not os.path.exists(dbnd_airflow_home):
            continue

        dbnd_log_init_msg(
            "Found airflow home folder at DBND, setting AIRFLOW_HOME to %s"
            % dbnd_airflow_home
        )
        os.environ[ENV_AIRFLOW_HOME] = dbnd_airflow_home



# © Copyright Databand.ai, an IBM Company 2022

import logging
import typing

from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import List

from dbnd._core.errors import DatabandRuntimeError
from dbnd._core.task_build.task_context import TaskContextPhase
from dbnd._core.task_build.task_registry import build_task_from_config
from dbnd._core.task_ctrl.task_dag import _TaskDagNode, all_subdags
from dbnd._core.task_run.task_run import TaskRun
from dbnd._core.utils.task_utils import (
    calculate_friendly_task_ids,
    tasks_summary,
    tasks_to_ids_set,
)
from dbnd_run.run_settings import EngineConfig, RunConfig


if typing.TYPE_CHECKING:
    from dbnd_run.run_executor.run_executor import RunExecutor
    from dbnd_run.task.task import Task

logger = logging.getLogger(__name__)


def check_if_completed_dfs(task, existing_completed_status=None):
    completed_status = existing_completed_status or dict()
    task_id = task.task_id
    completed_status[task_id] = completed = task._complete()
    if not completed:
        for c in task.ctrl.task_dag.upstream:
            if c.task_id in completed_status:
                continue
            completed_status.update(check_if_completed_dfs(c, completed_status))
    return completed_status


def check_if_completed_bfs(root_task, number_of_threads):
    completed_status = {}
    tasks_to_check_list = [root_task]

    with ThreadPoolExecutor(max_workers=number_of_threads) as executor:
        while tasks_to_check_list:
            new_task_to_check_list = []

            task_results = {}
            for task in tasks_to_check_list:
                task_results[executor.submit(task._complete)] = task.task_id

            for future in as_completed(task_results):
                task_id = task_results[future]
                try:
                    data = future.result()
                    completed_status[task_id] = data
                except Exception as e:
                    raise DatabandRuntimeError(
                        "Failed to get completeness result of task_id {}".format(
                            task_id
                        ),
                        nested_exceptions=e,
                    )

            for task in tasks_to_check_list:
                if completed_status[task.task_id]:
                    continue

                for upstream_task in task.ctrl.task_dag.upstream:
                    if upstream_task.task_id not in completed_status:
                        new_task_to_check_list.append(upstream_task)

            tasks_to_check_list = new_task_to_check_list

    return completed_status


def find_tasks_to_skip_complete(root_tasks, all_tasks, number_of_threads):
    # if True = should run, if False or None - should notcheck_if_completed
    completed_status = {}

    logger.info("Looking for completed tasks..")

    if number_of_threads > 1:
        for t in root_tasks:
            completed_status.update(check_if_completed_bfs(t, number_of_threads))
    else:
        for t in root_tasks:
            completed_status.update(check_if_completed_dfs(t))

    # only if completed_status is False task is not skipped
    # otherwise - it wasn't discovered or it's completed
    # it task is not discovered - we don't need to run it.

    skipped_tasks = {
        task for task in all_tasks if completed_status.get(task.task_id) is None
    }
    completed_tasks = {
        task for task in all_tasks if completed_status.get(task.task_id) is True
    }
    logger.info(
        "Found %s completed and %s skipped tasks",
        len(completed_tasks),
        len(skipped_tasks),
    )
    return completed_tasks, skipped_tasks


class TaskRunsBuilder(object):
    def get_tasks_without_disabled(self, task):
        td = task.ctrl.task_dag  # type: _TaskDagNode

        # we have to run twice as not runnable can have runnable upstream
        # TODO: validate that
        all_tasks = td.subdag_tasks()
        runnable_tasks = td.subdag_tasks(should_run_only=True)

        tasks_disabled = all_tasks.difference(runnable_tasks)

        return runnable_tasks, tasks_disabled

    def build_task_runs(self, run_executor, root_task, task_run_engine):
        # type: (RunExecutor, Task, EngineConfig) -> List[TaskRun]
        run_config = run_executor.run_config  # type: RunConfig

        # first, let remove all tasks explicitly marked as disabled by user
        tasks_to_run, tasks_disabled = self.get_tasks_without_disabled(root_task)
        if tasks_disabled:
            logger.info(
                "Tasks were removed from the task graph as they are marked as not to run: %s",
                tasks_summary(tasks_disabled),
            )

        roots = [root_task]
        tasks_skipped = set()
        # in case we need to run only part of the graph we mark all other tasks as skipped
        if run_config.selected_tasks_regex:
            task_dag = root_task.ctrl.task_dag  # type: _TaskDagNode
            roots = task_dag.select_by_task_names(
                run_config.selected_tasks_regex, tasks=tasks_to_run
            )

            tasks_skipped = tasks_to_run.difference(all_subdags(roots))

        enabled_tasks = tasks_to_run.difference(tasks_skipped)

        tasks_completed = set()
        task_skipped_as_not_required = set()
        if run_config.skip_completed:
            tasks_completed, task_skipped_as_not_required = find_tasks_to_skip_complete(
                roots, enabled_tasks, run_config.task_complete_parallelism_level
            )
        logger.debug("Completed tasks: %s", tasks_completed)
        # # if any of the tasks is spark add policy
        # from dbnd._core.task.spark import _BaseSparkTask
        # for t in tasks_to_run:
        #     if isinstance(t, _BaseSparkTask):
        #         t.spark.apply_spark_cluster_policy(t)
        # bash_op = BashOperator(task_id="echo", bash_command="echo hi")
        # self.root_task.set_upstream(bash_op.task)

        friendly_ids = calculate_friendly_task_ids(tasks_to_run)

        completed_ids = tasks_to_ids_set(tasks_completed)
        task_skipped_as_not_required_ids = tasks_to_ids_set(
            task_skipped_as_not_required
        )
        skipped_ids = tasks_to_ids_set(tasks_skipped)

        task_runs = []
        # should be sorted to have reproducible runs for snapshot testing
        for task in sorted(tasks_to_run, key=lambda x: x.task_creation_id):

            with task.ctrl.task_context(phase=TaskContextPhase.BUILD):
                # we want to have engine configuration with task overrides
                task_engine = build_task_from_config(
                    task_name=task_run_engine.task_name
                )
                task_engine.require_submit = task_run_engine.require_submit

                task_af_id = friendly_ids[task.task_id]
                task_run = run_executor.run.build_task_run(
                    task=task, task_af_id=task_af_id
                )
                run_executor.build_task_run_executor(task_run, task_engine)

            if task.task_id in completed_ids:
                logger.debug("Task is reused: %s : %s", task.task_id, task.task_outputs)
                task_run.is_reused = True

            if task.task_id in task_skipped_as_not_required_ids:
                logger.debug(
                    "Task is task_skipped_as_not_required_ids: %s : %s",
                    task.task_id,
                    task.task_outputs,
                )
                task_run.is_reused = True
                task_run.is_skipped_as_not_required = True

            if task.task_id in skipped_ids:
                task_run.is_skipped = True

            if task.task_id == root_task.task_id:
                task_run.is_root = True

            task_runs.append(task_run)

        return task_runs



# © Copyright Databand.ai, an IBM Company 2022

from typing import Any, Dict, Optional

from targets import FileTarget, target
from targets.errors import FileNotFoundException


class RunResultBand(object):
    """Wrap the access to the result of the run"""

    def __init__(self, paths):
        # type: (Dict[str, str])->RunResultBand
        self.paths = paths

    def load(self, name, value_type=None):
        # type: (str, Optional[str]) -> Any
        return target(self.paths[name]).load(value_type)

    def __iter__(self):
        return iter(self.paths)

    @classmethod
    def from_target(cls, results_map_target):
        # type: (FileTarget)->RunResultBand
        if results_map_target.exists():
            # we accepting the task_band of the pipeline:
            #   a json with a mapping between any result's name and its target
            paths = results_map_target.load(Dict[str, str])
            return cls(paths=paths)

        raise FileNotFoundException(
            "Couldn't access tasks results file at {location}.".format(
                location=results_map_target.path
            )
        )



# © Copyright Databand.ai, an IBM Company 2022

import getpass

from airflow import DAG, models, settings

from dbnd_run.airflow.compat import AIRFLOW_VERSION_2


def safe_get_context_manager_dag():
    """
    Try to find the CONTEXT_MANAGER_DAG object inside airflow.
    It was moved between versions, so we look for it in all the hiding places that we know of.
    """
    if AIRFLOW_VERSION_2:
        from airflow.models.dag import DagContext

        return DagContext.get_current_dag()

    if hasattr(settings, "CONTEXT_MANAGER_DAG"):
        return settings.CONTEXT_MANAGER_DAG
    elif hasattr(DAG, "_CONTEXT_MANAGER_DAG"):
        return DAG._CONTEXT_MANAGER_DAG
    elif hasattr(models, "_CONTEXT_MANAGER_DAG"):
        return models._CONTEXT_MANAGER_DAG

    return None


def hostname_as_username():
    """
    This function is used to speed up airflow.
    When not redirected to this method, airflow retrieves the host name of the machine using the `socket` library.
    The calls to `socket` are extremely slow, so we redirect it to the native python implementation
    """
    return getpass.getuser()



# © Copyright Databand.ai, an IBM Company 2022

import logging
import typing

from typing import List

from dbnd._core.constants import TaskExecutorType
from dbnd._core.context.use_dbnd_run import is_dbnd_orchestration_via_airflow_enabled
from dbnd._core.errors import DatabandConfigError
from dbnd_run import errors
from dbnd_run.plugin.dbnd_plugins import is_plugin_enabled
from dbnd_run.run_executor_engine.local_task_executor import LocalTaskExecutor


if typing.TYPE_CHECKING:

    from dbnd._core.task_run.task_run import TaskRun
    from dbnd_run.run_executor.run_executor import RunExecutor
    from dbnd_run.run_executor_engine import RunExecutorEngine
    from dbnd_run.run_settings import EngineConfig

logger = logging.getLogger(__name__)


def calculate_task_executor_type(submit_tasks, remote_engine, run_config):
    parallel = run_config.parallel
    task_executor_type = run_config.task_executor_type

    dbnd_run_airflow_enabled = is_dbnd_orchestration_via_airflow_enabled()

    if task_executor_type is None:
        if dbnd_run_airflow_enabled:
            from dbnd_run.airflow.executors import AirflowTaskExecutorType

            task_executor_type = AirflowTaskExecutorType.airflow_inprocess
        else:
            task_executor_type = TaskExecutorType.local

    if dbnd_run_airflow_enabled:
        from dbnd_run.airflow.executors import AirflowTaskExecutorType

        if parallel:
            if task_executor_type == TaskExecutorType.local:
                logger.warning(
                    "Auto switching to engine type '%s' due to parallel mode.",
                    AirflowTaskExecutorType.airflow_multiprocess_local,
                )
                task_executor_type = AirflowTaskExecutorType.airflow_multiprocess_local

            if task_executor_type == AirflowTaskExecutorType.airflow_inprocess:
                logger.warning(
                    "Auto switching to engine type '%s' due to parallel mode.",
                    AirflowTaskExecutorType.airflow_multiprocess_local,
                )
                task_executor_type = AirflowTaskExecutorType.airflow_multiprocess_local

        if (
            task_executor_type == AirflowTaskExecutorType.airflow_multiprocess_local
            or task_executor_type == AirflowTaskExecutorType.airflow_kubernetes
        ):
            from dbnd_run.airflow.db_utils import (
                airflow_sql_conn_url,
                airlow_sql_alchemy_conn,
            )

            if "sqlite" in airlow_sql_alchemy_conn():
                if run_config.enable_concurent_sqlite:
                    logger.warning(
                        "You are running parallel execution on top of sqlite database at %s! (see run.enable_concurent_sqlite)",
                        airflow_sql_conn_url(),
                    )
                else:
                    # in theory sqlite can support a decent amount of parallelism, but in practice
                    # the way airflow works each process holds the db exlusively locked which leads
                    # to sqlite DB is locked exceptions
                    raise errors.execute_engine.parallel_or_remote_sqlite(
                        task_executor_type
                    )

        if is_plugin_enabled("dbnd-docker"):
            from dbnd_docker.kubernetes.kubernetes_engine_config import (
                KubernetesEngineConfig,
            )

            if (
                submit_tasks
                and isinstance(remote_engine, KubernetesEngineConfig)
                and run_config.enable_airflow_kubernetes
            ):
                if task_executor_type != AirflowTaskExecutorType.airflow_kubernetes:
                    logger.info("Using dedicated kubernetes executor for this run")
                    task_executor_type = AirflowTaskExecutorType.airflow_kubernetes
                    parallel = True
    else:
        if parallel:
            logger.warning("Airflow is not installed, parallel mode is not supported")

    all_executor_types = [TaskExecutorType.local]
    if dbnd_run_airflow_enabled:
        from dbnd_run.airflow.executors import AirflowTaskExecutorType

        all_executor_types.extend(AirflowTaskExecutorType.all())

    if task_executor_type not in all_executor_types:
        raise DatabandConfigError(
            f"Unsupported engine type '{task_executor_type}' (supported: {all_executor_types})"
        )

    return task_executor_type, parallel


def get_task_executor(
    run_executor: "RunExecutor",
    task_executor_type: str,
    host_engine: "EngineConfig",
    target_engine: "EngineConfig",
    task_runs: "List[TaskRun]",
) -> "RunExecutorEngine":

    if task_executor_type == TaskExecutorType.local:
        return LocalTaskExecutor(
            run_executor,
            task_executor_type=task_executor_type,
            host_engine=host_engine,
            target_engine=target_engine,
            task_runs=task_runs,
        )
    else:

        from dbnd_run.airflow.dbnd_task_executor.dbnd_task_executor_via_airflow import (
            AirflowTaskExecutor,
        )

        return AirflowTaskExecutor(
            run_executor,
            task_executor_type=task_executor_type,
            host_engine=host_engine,
            target_engine=target_engine,
            task_runs=task_runs,
        )



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd._core.parameter.parameter_builder import output, parameter
from dbnd_run.tasks import PipelineTask, PythonTask


logger = logging.getLogger(__name__)


class SimplestTask(PythonTask):
    simplest_param = parameter.value("1")
    simplest_output = output

    def run(self):
        logger.info("We are running some simplest code!")
        self.simplest_output.write(self.simplest_param)


class SimplestPipeline(PipelineTask):
    simplest_output = output

    def band(self):
        self.simplest_output = SimplestTask().simplest_output


class SimpleTask(PythonTask):
    simple_input = parameter.data
    simple_output = output

    def run(self):
        v = self.simple_input.read()

        logger.info("We are running some simple code: got string of length %s", len(v))
        self.simple_output.write(v)



# © Copyright Databand.ai, an IBM Company 2022

import logging
import os

from dbnd._core.cli.click_utils import _help
from dbnd._core.configuration.environ_config import is_unit_test_mode
from dbnd._core.utils.project.project_fs import databand_system_path
from dbnd._vendor.click._bashcomplete import ZSH_COMPLETION_SUFFIX
from dbnd._vendor.click.utils import make_default_short_help
from dbnd_run.cli.utils import FastSingletonContext, PrefixStore, no_errors


logger = logging.getLogger(__name__)

# indices
TASK_FAMILY = "task_by_family"
TASK_FULL = "task_by_full"
CONFIG_FAMILY = "config_by_family"
CONFIG_FULL = "config_by_full"
PARAM_FAMILY = "param_by_family"
PARAM_FULL = "param_by_full"

# blacklists
PARAM_BLACKLIST = ["task_env", "task_target_date"]


class AutoCompleter(FastSingletonContext):
    def __init__(self):
        self._store_ins = None

    def config(self):
        return self._build_task_completer(is_config=True)

    def task(self):
        return self._build_task_completer(is_config=False)

    def config_param(self):
        return self._build_param_completer(
            is_config=True, with_root=False, with_tasks=True
        )

    def task_param(self):
        return self._build_param_completer(
            is_config=False, with_root=True, with_tasks=True
        )

    def root_param(self):
        return self._build_param_completer(
            is_config=False, with_root=True, with_tasks=False
        )

    @no_errors()
    def refresh(self, task_classes):
        from dbnd._core.parameter.parameter_definition import _ParameterKind
        from dbnd._core.task.config import Config

        store = self._get_store(loaded=False)

        for task_cls in task_classes:
            if task_cls is Config:
                continue

            task_family = task_cls.get_task_family()
            full_task_family = task_cls.task_definition.full_task_family
            is_config = issubclass(task_cls, Config)
            description = task_cls.__doc__

            store.set(full_task_family, description)

            if is_config:
                store.index(CONFIG_FAMILY, task_family, full_task_family)
                store.index(CONFIG_FULL, full_task_family, full_task_family)
            else:
                store.index(TASK_FAMILY, task_family, full_task_family)
                store.index(TASK_FULL, full_task_family, full_task_family)

            for (
                param_name,
                param_obj,
            ) in task_cls.task_definition.task_param_defs.items():
                if (
                    param_obj.system
                    or param_obj.kind == _ParameterKind.task_output
                    or param_name in PARAM_BLACKLIST
                ):
                    continue

                param_family = task_family + "." + param_name
                param_full = full_task_family + "." + param_name

                store.set(param_full, param_obj.description)
                store.index(PARAM_FAMILY, param_family, param_full)
                store.index(PARAM_FULL, param_full, param_full)

        store.save()

    def _build_task_completer(self, is_config):
        indices = _get_indices(is_config)

        @no_errors(ret=tuple())
        def _completer(ctx, args, incomplete):
            store = self._get_store(loaded=True)
            tasks = store.search(indices, incomplete)

            if incomplete in tasks:
                return []

            res = []
            for task_name, task_desc in tasks.items():
                short_desc = _help(task_desc)
                res.append((task_name, short_desc))

            return res

        return _completer

    def _build_param_completer(self, is_config, with_root, with_tasks):
        indices = _get_indices(is_config)

        @no_errors(ret=tuple())
        def _completer(ctx, args, incomplete):
            store = self._get_store(loaded=True)
            res = [(ZSH_COMPLETION_SUFFIX, "")]

            if with_root and len(ctx.args) > 0:
                params = store.search(
                    indices_names=[PARAM_FAMILY, PARAM_FULL],
                    prefix=ctx.args[0] + "." + incomplete,
                )

                for param_name, param_desc in params.items():
                    _, param_name = param_name.rsplit(".", maxsplit=1)
                    short_desc = _help(param_desc)
                    res.append((param_name, short_desc))

            if not with_tasks:
                return res

            task_name = incomplete
            if "." in incomplete:
                task_name, _ = incomplete.rsplit(".", maxsplit=1)

            tasks = store.search(indices, task_name)

            if task_name in tasks:
                params = store.search(
                    indices_names=[PARAM_FAMILY, PARAM_FULL], prefix=incomplete
                )

                for key, param_desc in params.items():
                    short_desc = _help(param_desc)
                    res.append((key, short_desc))

                return res

            for task_name, task_desc in tasks.items():
                short_desc = _help_short(task_desc)
                res.append((task_name + ".", short_desc))

            return res

        return _completer

    def _get_store(self, loaded):
        if self._store_ins is None:

            autocompletion_cache_filename = (
                ".shell.unittest" if is_unit_test_mode() else ".shell"
            )
            store_path = databand_system_path(
                "autocompletion", autocompletion_cache_filename
            )

            self._store_ins = PrefixStore(store_path)

            if loaded and os.path.exists(store_path):
                self._store_ins.load()

        return self._store_ins


def _get_indices(is_config):
    if is_config:
        return [CONFIG_FAMILY, CONFIG_FULL]
    return [TASK_FAMILY, TASK_FULL]


def _help_short(string):
    return make_default_short_help(string or "-")


completer = AutoCompleter.try_instance()



# © Copyright Databand.ai, an IBM Company 2022


def show_run_url(run_url):
    import IPython

    html = 'Databand Run is <a href="%s" target="_blank" >at this link</a>' % run_url
    IPython.display.display(IPython.display.HTML(html))



# © Copyright Databand.ai, an IBM Company 2022

# ORIGIN: https://github.com/databricks/mlflow : mlflow/store/tracking_store_file.py
from __future__ import print_function

import json
import logging
import os
import re
import time
import typing

from datetime import datetime

import six
import yaml

from six import BytesIO

from dbnd._core.constants import MetricSource, TaskRunState
from dbnd._core.errors import DatabandError, DatabandRuntimeError
from dbnd._core.task_run.task_run_meta_files import TaskRunMetaFiles
from dbnd._core.tracking.backends import TrackingStore
from dbnd._core.tracking.schemas.metrics import Artifact, Metric
from dbnd._core.tracking.tracking_info_convertor import (
    build_task_run_info,
    task_to_task_def,
)
from dbnd.api.serialization.run import RunInfoSchema
from dbnd.api.serialization.task import TaskDefinitionInfoSchema, TaskRunInfoSchema
from targets import target


if typing.TYPE_CHECKING:
    from typing import Iterable, List

    from dbnd._core.task_run.task_run import TaskRun
    from targets.value_meta import ValueMeta


logger = logging.getLogger(__name__)


def _parse_metric(value):
    try:
        return float(value)
    except ValueError:
        return value


_METRICS_RE = re.compile(r"(\d+)\s+(.+)")


class FileTrackingStore(TrackingStore):
    def set_task_run_state(self, task_run, state, error=None, timestamp=None):
        if state == TaskRunState.RUNNING:
            self.dump_task_run_info(task_run)

    def _get_meta_files(self, task_run):
        return task_run.task_run_executor.meta_files

    def dump_task_run_info(self, task_run):

        info = {
            "task_run_executor": TaskRunInfoSchema()
            .dump(build_task_run_info(task_run))
            .data,
            "task_def": (
                TaskDefinitionInfoSchema()
                .dump(task_to_task_def(task_run.run.context, task_run.task))
                .data
            ),
            "run": RunInfoSchema().dump(task_run.run).data,
        }
        # old implementation, replace with "dbnd_tracking" request

        meta_data_file = self._get_meta_files(task_run).get_meta_data_file()
        with meta_data_file.open("w") as yaml_file:
            yaml.dump(info, yaml_file, default_flow_style=False)

    def log_histograms(self, task_run, key, value_meta, timestamp):
        # type: (TaskRun, str, ValueMeta, datetime) -> None
        metric_path = self._get_meta_files(task_run).get_metric_target(
            "{}.json".format(key), source=MetricSource.histograms
        )
        data_schema = value_meta.data_schema.as_dict() if value_meta.data_schema else {}
        data = json.dumps(
            {
                "timestamp": int(timestamp.timestamp()),
                "metrics": {
                    "schema": data_schema,
                    "preview": value_meta.value_preview,
                    "shape": value_meta.data_dimensions,
                    "stats": value_meta._get_stats_dict_from_columns_stats(),
                    "histograms": value_meta.histograms,
                },
            }
        )
        metric_path.write(data)

    def log_metrics(self, task_run, metrics):
        # type: (TaskRun, List[Metric]) -> None
        for metric in metrics:
            metric_path = self._get_meta_files(task_run).get_metric_target(
                metric.key, source=metric.source
            )
            timestamp = int(time.mktime(metric.timestamp.timetuple()))
            value = "{} {}\n".format(timestamp, metric.serialized_value)

            data = value
            if metric_path.exists():
                data = metric_path.read() + value
            metric_path.write(data)

    def log_artifact(self, task_run, name, artifact, artifact_target):
        artifact_target.mkdir_parent()

        if isinstance(artifact, six.string_types):
            from targets.dir_target import DirTarget

            artifact_target_source = target(artifact)
            if isinstance(artifact_target_source, DirTarget):
                artifact_target_source.copy(artifact_target)
            else:
                data = artifact_target_source.read()
                artifact_target.write(data)

            return artifact_target

        try:
            from matplotlib.figure import Figure

            if isinstance(artifact, Figure):
                temp = BytesIO()
                artifact.savefig(temp)
                temp.seek(0)
                artifact_target.write(temp.read(), mode="wb")
                return artifact_target
        except ImportError:
            pass

        raise DatabandRuntimeError(
            "Could not recognize artifact of type %s, must be string or matplotlib Figure"
            % type(artifact)
        )

    def is_ready(self):
        return True


class TaskRunMetricsFileStoreReader(object):
    def __init__(self, attempt_folder, **kwargs):
        super(TaskRunMetricsFileStoreReader, self).__init__(**kwargs)
        self.meta = TaskRunMetaFiles(attempt_folder)

    def _get_all_metrics_names(self, source=None):
        metrics_root = self.meta.get_metric_folder(source=source)
        all_files = [os.path.basename(str(p)) for p in metrics_root.list_partitions()]
        return [re.sub(r"\.json\b", "", f) for f in all_files]

    def get_metric_history(self, key, source=None):
        metric_target = self.meta.get_metric_target(key, source=source)
        if not metric_target.exists():
            raise DatabandError("Metric '%s' not found" % key)
        metric_data = metric_target.readlines()
        rsl = []
        for pair in metric_data:
            ts, val = pair.strip().split(" ")
            rsl.append(Metric(key, float(val), datetime.fromtimestamp(int(ts))))
        return rsl

    def get_all_metrics_values(self, source=None):
        metrics = []
        for key in self._get_all_metrics_names(source=source):
            try:
                metrics.extend(self.get_metrics(key, source=source))
            except Exception as ex:
                raise DatabandError(
                    "Failed to read metrics for %s at %s" % (key, self.meta.root),
                    nested_exceptions=ex,
                )
        return {m.key: m.value for m in metrics}

    def get_run_info(self):
        with self.meta.get_meta_data_file().open("r") as yaml_file:
            return RunInfoSchema().load(**yaml.load(yaml_file))

    def get_metrics(self, key, source=None):
        # type: (str, MetricSource) -> Iterable[Metric]
        if source == MetricSource.histograms:
            return self.get_histogram_metrics(key)

        metric_target = self.meta.get_metric_target(key, source=source)
        if not metric_target.exists():
            raise DatabandRuntimeError("Metric '%s' not found" % key)
        metric_data = metric_target.readlines()
        if len(metric_data) == 0:
            raise DatabandRuntimeError("Metric '%s' is malformed. No data found." % key)
        first_line = metric_data[0]

        metric_parsed = _METRICS_RE.match(first_line)
        if not metric_parsed:
            raise DatabandRuntimeError(
                "Metric '%s' is malformed. Expected format: 'TS VALUE', got='%s'"
                % (key, first_line)
            )

        timestamp, val = metric_parsed.groups()

        metric = Metric(
            key=key,
            value=_parse_metric(val),
            timestamp=datetime.fromtimestamp(int(timestamp)),
        )
        return [metric]

    def get_histogram_metrics(self, key):
        # type: (str) -> Iterable[Metric]
        metric_target = self.meta.get_metric_target(
            "{}.json".format(key), source=MetricSource.histograms
        )
        hist_metrics = json.load(metric_target)
        timestamp = hist_metrics["timestamp"]
        metrics = hist_metrics["metrics"]
        for name, value in metrics.items():
            if not isinstance(value, (dict, list)):
                yield Metric(
                    key="{}.{}".format(key, name),
                    value=value,
                    timestamp=datetime.fromtimestamp(timestamp),
                )
                continue

            yield Metric(
                key="{}.{}".format(key, name),
                value_json=value,
                timestamp=datetime.fromtimestamp(timestamp),
            )
            if name == "stats":
                for column, stats in value.items():
                    for stat, val in stats.items():
                        yield Metric(
                            key="{}.{}.{}".format(key, column, stat),
                            value=val,
                            timestamp=datetime.fromtimestamp(timestamp),
                        )
            elif name == "shape":
                for dim, val in enumerate(value):
                    yield Metric(
                        key="{}.shape{}".format(key, dim),
                        value=val,
                        timestamp=datetime.fromtimestamp(timestamp),
                    )

    def get_artifact(self, name):
        artifact_target = self.meta.get_artifact_target(name)
        if not artifact_target.exists():
            raise DatabandError("Artifact '%s' not found" % name)
        return Artifact(artifact_target.path)


def read_task_metrics(attempt_folder, source=None):
    return TaskRunMetricsFileStoreReader(
        attempt_folder=target(attempt_folder)
    ).get_all_metrics_values(source=source)



# © Copyright Databand.ai, an IBM Company 2022

import contextlib
import logging
import os
import signal
import subprocess
import sys
import typing

from time import sleep, time

from dbnd._core.configuration.environ_config import ENV_DBND__ORCHESTRATION__NO_PLUGINS
from dbnd._core.constants import RunState
from dbnd._core.tracking.backends import TrackingStore
from dbnd._core.utils.basics.format_exception import format_exception_as_str
from dbnd._vendor.psutil.vendorized_psutil import pid_exists


if typing.TYPE_CHECKING:
    from dbnd_run.run_executor.run_executor import RunExecutor

logger = logging.getLogger(__name__)

TERMINATE_WAIT_TIMEOUT = 5


@contextlib.contextmanager
def start_heartbeat_sender(run_executor):
    # type: (RunExecutor)-> ...
    """
    Context that will run hearbeat sender on __enter__
    """
    core = run_executor.run.context.settings.core
    run_config = run_executor.run_config
    heartbeat_interval_s = run_config.heartbeat_interval_s

    if not heartbeat_interval_s > 0:
        logger.info(
            "run heartbeat sender disabled (set task.heartbeat_interval_s to value > 0)"
        )
        yield
        return

    if not core.tracker_api:
        logger.info("run heartbeat sender disabled (set core.tracker_api)")
        yield
        return

    sp = None
    heartbeat_log_fp = None
    try:
        try:
            cmd = [
                sys.executable,
                "-m",
                "dbnd",
                "send-heartbeat",
                "--run-uid",
                str(run_executor.run.run_uid),
                "--driver-pid",
                str(os.getpid()),
                "--heartbeat-interval",
                str(heartbeat_interval_s),
                "--tracker",
                ",".join(core.tracker),
                "--tracker-api",
                core.tracker_api,
            ]
            if core.databand_url:
                cmd += ["--databand-url", core.databand_url]

            if run_config.heartbeat_sender_log_to_file:
                local_heartbeat_log_file = run_executor.run_local_root.partition(
                    name="heartbeat.log"
                )
                heartbeat_log_file = local_heartbeat_log_file
                heartbeat_log_fp = heartbeat_log_file.open("w")
                stdout = heartbeat_log_fp
                logger.info(
                    "Starting heartbeat with log at %s using cmd: %s",
                    heartbeat_log_file,
                    subprocess.list2cmdline(cmd),
                )
            else:
                stdout = None
                logger.info(
                    "Starting heartbeat using cmd: %s", subprocess.list2cmdline(cmd)
                )
            env = os.environ.copy()
            if run_config.hearbeat_disable_plugins:
                # we might disable plugins, as underline process doesn't need that
                env[ENV_DBND__ORCHESTRATION__NO_PLUGINS] = "True"

            sp = subprocess.Popen(cmd, stdout=stdout, stderr=subprocess.STDOUT, env=env)
        except Exception as ex:
            logger.info(
                "Failed to spawn heartbeat process, you can disable it via [task]heartbeat_interval_s=0  .\n %s",
                ex,
            )
            raise ex
        yield
    finally:
        if sp:
            sp.terminate()

            try:
                sp.wait(timeout=TERMINATE_WAIT_TIMEOUT)
            except Exception:
                logger.warning(
                    "waited %s seconds for the heartbeat sender to exit but it still hasn't exited",
                    TERMINATE_WAIT_TIMEOUT,
                )

        if heartbeat_log_fp:
            heartbeat_log_fp.close()


def send_heartbeat_continuously(
    run_uid, tracking_store, heartbeat_interval_s, driver_pid
):  # type: (str, TrackingStore, int, int) -> None
    logger.info(
        "[heartbeat sender] starting heartbeat sender process (pid %s) with a send interval of %s seconds"
        % (os.getpid(), heartbeat_interval_s)
    )

    try:
        while True:
            loop_start = time()
            try:
                if not pid_exists(
                    driver_pid
                ):  # failsafe, in case the driver process died violently
                    logger.info(
                        "[heartbeat sender] driver process %s stopped, stopping heartbeat sender",
                        driver_pid,
                    )
                    return

                run_state = tracking_store.heartbeat(run_uid=run_uid)
                logger.debug("[heartbeat sender] sent heartbeat")
                if run_state == RunState.SHUTDOWN.value:
                    logger.info(
                        "[heartbeat sender] received run state SHUTDOWN: killing driver process"
                    )
                    os.kill(driver_pid, signal.SIGTERM)
            except KeyboardInterrupt:
                logger.info(
                    "[heartbeat sender] stopping heartbeat sender process due to interrupt"
                )
                return
            except Exception:
                logger.warning(
                    "[heartbeat sender] failed to send heartbeat: %s",
                    format_exception_as_str(),
                )

            time_to_sleep_s = max(0, time() + heartbeat_interval_s - loop_start)
            if time_to_sleep_s > 0:
                sleep(time_to_sleep_s)
    except KeyboardInterrupt:
        return
    except Exception:
        logger.exception("[heartbeat sender] Failed to run heartbeat")
    finally:
        logger.info("[heartbeat sender] stopping heartbeat sender")
        sys.exit(0)



# © Copyright Databand.ai, an IBM Company 2022

from __future__ import print_function

import logging
import os
import sys

from logging.config import DictConfigurator
from typing import Callable, List, Optional

from dbnd._core.log.config import configure_logging_dictConfig
from dbnd._core.log.logging_utils import (
    find_handler,
    safe_is_typeof,
    setup_log_file,
    try_init_sentry,
)
from dbnd._core.parameter.parameter_builder import parameter
from dbnd._core.task import config
from dbnd._core.utils.project.project_fs import databand_system_path


logger = logging.getLogger(__name__)


class RunLoggingConfig(config.Config):
    """Databand's logger configuration"""

    _conf__task_family = "run_log"
    disabled = parameter(
        description="Determine whether logging should be disabled."
    ).value(False)
    debug_log_config = parameter(
        description="Enable debugging our logging configuration system."
    ).value(False)

    override_airflow_logging_on_task_run = parameter(
        description="Enable replacing airflow logger with databand logger."
    ).value(True)
    support_jupyter = parameter(
        description="Support logging output to Jupiter UI."
    ).value(False)

    level = parameter(
        description="Set which logging level will be used. This could be DEBUG, INFO, WARN, or ERROR"
    ).value("INFO")
    formatter = parameter(
        description="Set the log formatting string, using the logging library convention."
    )[str]
    formatter_colorlog = parameter(
        description="Set the log formatting string, using the logging library convention."
    )[str]
    formatter_simple = parameter(
        description="Set the log formatting string, using the logging library convention."
    )[str]

    console_formatter_name = parameter(
        description="Set the name of the formatter logging to console output."
    )[str]
    file_formatter_name = parameter(
        description="Set the name of the formatter logging to file output."
    )[str]

    # sentry config
    sentry_url = parameter(
        default=None,
        description="Set the URL for setting up sentry logger. Make sure the url is exposed to dbnd run environment",
    )[str]
    sentry_env = parameter(
        default="dev", description="Set the environment for sentry logger."
    )[str]
    sentry_release = parameter(default="", description="Release for sentry logger")[str]
    sentry_debug = parameter(default=False, description="Enable debug flag for sentry")[
        bool
    ]

    file_log = parameter(
        default=None,
        description="Determine whether logger should log to a file. This is off by default",
    )[str]

    stream_stdout = parameter(
        description="Should Databand's logger stream stdout instead of stderr."
    ).value(False)

    custom_dict_config = parameter(
        default=None, description="Set customized logging configuration."
    )[Callable]

    at_warn = parameter.help("Set name of loggers to put in WARNING mode.").c[List[str]]
    at_debug = parameter.help("Set name of loggers to put in DEBUG mode.").c[List[str]]

    capture_task_run_log = parameter(
        default=True, description="Enable capturing task output into log."
    )

    send_body_to_server = parameter(
        default=True, description="Enable or disable sending log file to server."
    )[bool]

    remote_logging_disabled = parameter.help(
        "For tasks using a cloud environment, don't copy the task log to cloud storage."
    ).value(False)

    targets_log_level = parameter(
        default="DEBUG",
        description="Should log the time it takes for marshalling and unmarshalling targets.",
    )[str]

    disable_colors = parameter(default=False, description="Disable any colored logs.")

    sqlalchemy_print = parameter(description="Enable sqlalchemy logger.").value(False)
    sqlalchemy_trace = parameter(
        description="Enable tracing sqlalchemy queries."
    ).value(False)

    def _initialize(self):
        super(RunLoggingConfig, self)._initialize()
        self.task_log_file_formatter = None

    def get_dbnd_logging_config(self, filename=None):
        if self.custom_dict_config:
            logger.info("Using user provided logging config")
            return self.settings.log.custom_dict_config()

        return self.get_dbnd_logging_config_base(filename=filename)

    def get_dbnd_logging_config_base(self, filename=None):
        # type: (RunLoggingConfig, Optional[str]) -> Optional[dict]
        self.log_debug("Using log.get_dbnd_logging_config_base")
        log_settings = self
        log_level = log_settings.level
        # we want to have "real" output, so nothing can catch our handler
        # in opposite to what airflow is doing
        console_stream = (
            sys.__stdout__ if log_settings.stream_stdout else sys.__stderr__
        )

        if "ipykernel" in sys.modules and self.support_jupyter:
            # we can not use __stdout__ or __stderr__ as it will not be printed into jupyter web UI
            # at the same time  using sys.stdout when airflow is active is very dangerous
            # as it can create dangerous loop from airflow redirection into root logger

            self.log_debug("ipykernel: checking on console_stream again")
            console_stream = sys.stdout if log_settings.stream_stdout else sys.stderr

        # dummy path, we will not write to this file
        task_file_handler_file = databand_system_path("logs", "task.log")

        self.log_debug("task_file_handler_file: %s", task_file_handler_file)
        setup_log_file(task_file_handler_file)

        config = {
            "version": 1,
            "disable_existing_loggers": False,
            "filters": {
                "task_context_filter": {
                    "()": "dbnd._core.log.logging_utils.TaskContextFilter"
                }
            },
            "formatters": {
                "formatter": {"format": log_settings.formatter},
                "formatter_simple": {"format": log_settings.formatter_simple},
                "formatter_colorlog": {
                    "()": "dbnd._vendor.colorlog.ColoredFormatter",
                    "format": log_settings.formatter_colorlog,
                    "reset": True,
                },
            },
            "handlers": {
                "console": {
                    "class": "logging.StreamHandler",
                    "stream": console_stream,
                    "formatter": log_settings.console_formatter_name,
                    "filters": ["task_context_filter"],
                }
            },
            "root": {"handlers": ["console"], "level": log_level},
        }
        if filename:
            setup_log_file(filename)
            config["handlers"]["file"] = {
                "class": "logging.FileHandler",
                "formatter": log_settings.file_formatter_name,
                "filename": filename,
                "encoding": "utf-8",
            }
            config["root"]["handlers"].append("file")

        loggers = config.setdefault("loggers", {})
        for logger_warn in log_settings.at_warn:
            loggers[logger_warn] = {"level": logging.WARNING, "propagate": True}

        for logger_debug in log_settings.at_debug:
            loggers[logger_debug] = {"level": logging.DEBUG, "propagate": True}

        if log_settings.sqlalchemy_print:
            loggers["sqlalchemy.engine"] = {"level": logging.INFO, "propagate": True}

        self.log_debug("Log config: %s", config)
        return config

    def configure_dbnd_logging(self):
        if self.disabled:
            self.log_debug("Log is disabled, skipping configure_dbnd_logging")
            return

        # start by trying to initiate Sentry setup - has side effect of changing the logging config
        self.log_debug("Initialize Sentry setup")
        try_init_sentry(self)

        if self.disable_colors:
            self.log_debug("Colors are disabled")
            self.disable_color_logs()

        dict_config = self.get_dbnd_logging_config(filename=self.file_log)

        airflow_task_log_handler = None
        if self.override_airflow_logging_on_task_run:
            airflow_task_log_handler = self.dbnd_override_airflow_logging_on_task_run()

        try:
            self.log_debug("configure_logging_dictConfig: %s", dict_config)
            configure_logging_dictConfig(dict_config=dict_config)
        except Exception as e:
            # we print it this way, as it could be that now "logging" is down!
            print(
                "Failed to load reload logging configuration with dbnd settings! Exception: %s"
                % (e,),
                file=sys.__stderr__,
            )
            raise
        if airflow_task_log_handler:
            self.log_debug("logging.root.handlers.append(airflow_task_log_handler)")
            logging.root.handlers.append(airflow_task_log_handler)
        self.log_debug("Databand logging is up!")

    def dbnd_override_airflow_logging_on_task_run(self):
        # EXISTING STATE:
        # root logger use Console handler -> prints to current sys.stdout
        # on `airflow run` without interactive -> we have `redirect_stderr` applied that will redirect sys.stdout
        # into logger `airflow.task`, that will save everything into file.
        #  EVERY output of root logger will go through CONSOLE handler into AIRFLOW.TASK without being printed to screen

        self.log_debug("dbnd_override_airflow_logging_on_task_run")
        if not sys.stderr or not safe_is_typeof(sys.stderr, "StreamLogWriter"):
            self.log_debug(
                "Airflow logging is already replaced by dbnd stream log writer! sys.stderr=%s",
                sys.stderr,
            )
            return

        # NEW STATE
        # we will move airflow.task file handler to root level
        # we will set propogate
        # we will stop redirect of airflow logging

        # this will disable stdout ,stderr redirection
        sys.stderr = sys.__stderr__
        sys.stdout = sys.__stdout__

        airflow_root_console_handler = find_handler(logging.root, "console")

        self.log_debug("airflow_root_console_handler:%s", airflow_root_console_handler)
        if safe_is_typeof(airflow_root_console_handler, "RedirectStdHandler"):
            # we are removing this console logger
            # this is the logger that capable to create self loop
            # as it writes to "latest" sys.stdout,
            # if you have stdout redirection into any of loggers, that will propogate into root
            # you get very busy message loop that is really hard to debug

            self.log_debug("airflow_root_console_handler has been removed")
            logging.root.handlers.remove(airflow_root_console_handler)

        airflow_task_logger = logging.getLogger("airflow.task")

        self.log_debug("airflow_task_logger: %s", airflow_task_logger)
        airflow_task_log_handler = find_handler(airflow_task_logger, "task")
        if airflow_task_log_handler:
            self.log_debug("airflow_task_log_handler: %s", airflow_task_log_handler)
            logging.root.handlers.append(airflow_task_log_handler)
            airflow_task_logger.propagate = True
            airflow_task_logger.handlers = []
        self.log_debug(
            "dbnd_override_airflow_logging_on_task_run logging.root: %s", logging.root
        )
        return airflow_task_log_handler

    def get_task_log_file_handler(self, log_file):
        if not self.task_log_file_formatter:
            config = self.get_dbnd_logging_config()
            configurator = DictConfigurator(config)
            file_formatter_config = configurator.config.get("formatters").get(
                self.file_formatter_name
            )
            self.task_log_file_formatter = configurator.configure_formatter(
                file_formatter_config
            )

        log_file = str(log_file)
        setup_log_file(log_file)
        handler = logging.FileHandler(filename=log_file, encoding="utf-8")
        handler.setFormatter(self.task_log_file_formatter)
        handler.setLevel(self.level)
        return handler

    def disable_color_logs(self):
        """Removes colors from any console related config"""
        logger.debug("disabling color logs")

        os.environ["ANSI_COLORS_DISABLED"] = "True"  # disabling termcolor.colored
        self.exception_no_color = True
        if self.console_formatter_name == "formatter_colorlog":
            self.console_formatter_name = "formatter_simple"

    def log_debug(self, msg, *args):
        if not self.debug_log_config:
            if not self.disabled:
                # we don't want to print ANYTHING if we are disabled
                logger.debug(msg, *args)
            return

        try:
            # we print to stderr as well in case logging is broken
            print("DEBUG_LOG_CONFIG:" + msg % args, file=sys.__stderr__)
            logger.info("DEBUG_LOG_CONFIG:" + msg, *args)
        except Exception:
            pass



# © Copyright Databand.ai, an IBM Company 2022

import contextlib
import logging
import os
import sys
import threading
import typing

from typing import Iterator, Optional, Union
from uuid import UUID

import six

from dbnd._core.configuration.dbnd_config import config
from dbnd._core.configuration.environ_config import (
    ENV_DBND__USER_PRE_INIT,
    is_unit_test_mode,
)
from dbnd._core.constants import (
    RunState,
    SystemTaskName,
    TaskExecutorType,
    TaskRunState,
)
from dbnd._core.context.use_dbnd_run import is_dbnd_orchestration_via_airflow_enabled
from dbnd._core.current import current_task_run
from dbnd._core.errors import DatabandRunError, DatabandSystemError
from dbnd._core.errors.base import (
    DatabandError,
    DatabandFailFastError,
    DatabandSigTermError,
    DbndCanceledRunError,
)
from dbnd._core.run.databand_run import DatabandRun
from dbnd._core.task_build.task_registry import get_task_registry
from dbnd._core.task_run.task_run import TaskRun
from dbnd._core.task_run.task_run_error import TaskRunError
from dbnd._core.tracking.schemas.tracking_info_run import ScheduledRunInfo
from dbnd._core.utils import console_utils
from dbnd._core.utils.basics.load_python_module import (
    load_python_callable,
    run_user_func,
)
from dbnd._core.utils.basics.nested_context import nested
from dbnd._core.utils.basics.singleton_context import SingletonContext
from dbnd._core.utils.pycharm_debugger import start_pycharm_debugger
from dbnd._core.utils.seven import cloudpickle
from dbnd._core.utils.task_utils import get_project_name_safe, get_task_name_safe
from dbnd._core.utils.timezone import utcnow
from dbnd._core.utils.traversing import flatten
from dbnd._core.utils.uid_utils import get_uuid
from dbnd.api.runs import kill_run
from dbnd_run.plugin.dbnd_plugins import pm
from dbnd_run.run_executor.factory import (
    calculate_task_executor_type,
    get_task_executor,
)
from dbnd_run.run_executor.heartbeat_sender import start_heartbeat_sender
from dbnd_run.run_executor.results_view import RunResultBand
from dbnd_run.run_executor.task_runs_builder import TaskRunsBuilder
from dbnd_run.run_executor_engine.local_task_executor import (
    LocalTaskExecutor,
    topological_sort,
)
from dbnd_run.run_settings import EnvConfig, RunLoggingConfig, RunSettings
from dbnd_run.run_settings.engine import build_engine_config
from dbnd_run.run_settings.run import RunConfig
from dbnd_run.task.task import Task
from dbnd_run.task_ctrl.task_dag_describe import print_tasks_tree
from dbnd_run.task_ctrl.task_run_executor import TaskRunExecutor
from targets import FileTarget, target
from targets.caching import TARGET_CACHE
from targets.providers.pandas import register_pd_to_hdf5_as_table_marshaler


if typing.TYPE_CHECKING:
    from dbnd._core.context.databand_context import DatabandContext

DEFAULT_TASK_CANCELED_ERR_MSG = "Task was killed by the user"

logger = logging.getLogger(__name__)

# naive implementation of stop event
# we can't save it on Context (non pickable in some cases like running in multithread python)
# if somebody is killing run it's global for the whole process
_is_killed = threading.Event()


class RunExecutor(SingletonContext):
    """
    This class is in charge of running the pipeline at orchestration (dbnd run) mode
    It wraps it's own execution with _RunExecutor_Task task, so logs and state are reported to tracker
    * in submitter mode - send same command to remote engine
    * in driver mode - build and run the task
    """

    def __init__(
        self, run, root_task_or_task_name, send_heartbeat, force_task_name=None
    ):

        self.run = run  # type: DatabandRun
        self.databand_context = run.context  # type: DatabandContext
        self.send_heartbeat = send_heartbeat

        if root_task_or_task_name is None:
            raise DatabandSystemError(
                "Run executor requires task name or task, got None"
            )

        # we are building it only in driver,
        # root task sometimes can be executed only inside the docker
        # it can be affected by docker env (k8s secrets/external source)
        self.root_task_name_to_build = None
        self.force_task_name = force_task_name
        if isinstance(root_task_or_task_name, six.string_types):
            self.root_task_name_to_build = root_task_or_task_name
        elif isinstance(root_task_or_task_name, Task):
            # we have a ready task, we will not build it, just run
            self.run.root_task = root_task_or_task_name
        else:
            raise DatabandSystemError(
                "Run executor requires task name or task, got %s - %s",
                root_task_or_task_name,
                type(root_task_or_task_name),
            )

        self.run_folder_prefix = os.path.join(
            "log",
            run.execution_date.strftime("%Y-%m-%d"),
            "%s_%s_%s"
            % (
                run.execution_date.strftime("%Y-%m-%dT%H%M%S.%f"),
                run.job_name,
                run.name,
            ),
        )
        self.run_root = self.env.dbnd_root.folder(self.run_folder_prefix)
        self.run_local_root = self.env.dbnd_local_root.folder(self.run_folder_prefix)

        self.driver_dump = self.run_root.file("run.pickle")

        self.local_engine = build_engine_config(self.env.local_engine)
        self.remote_engine = build_engine_config(
            self.env.remote_engine or self.env.local_engine
        )

        # we take values from run config (if defined) - otherwise from specific env definition
        # usually run_config will contain "negative" override
        # values at env_config are based on env.remote_config ( try to submit_driver/submit_tasks if defined)
        self.submit_driver = (
            self.run_config.submit_driver
            if self.run_config.submit_driver is not None
            else self.env.submit_driver
        )
        self.submit_tasks = (
            self.run_config.submit_tasks
            if self.run_config.submit_tasks is not None
            else self.env.submit_tasks
        )
        self.task_executor_type, self.parallel = calculate_task_executor_type(
            self.submit_tasks, self.remote_engine, self.run_config
        )

        self.local_engine = build_engine_config(self.env.local_engine).clone(
            require_submit=False
        )
        if self.submit_driver and not run.existing_run:
            # we are running submitter, that will send driver to remote
            self.run_executor_type = SystemTaskName.driver_submit
            self.host_engine = self.local_engine
        else:
            # We are running Driver ( submitter already sent this , or no submitter at all)
            self.run_executor_type = SystemTaskName.driver
            if self.submit_driver:
                # submit drive is true, but we are in existing run:
                # we are after the jump from submit to driver execution (to remote engine)
                self.host_engine = self.remote_engine
            else:
                self.host_engine = self.local_engine
            if not self.submit_tasks or self.task_executor_type == "airflow_kubernetes":
                # if we are not in submit tasks, we disable engine "resubmit"
                # airflow kubernetes - we don't want task resubmition, even if engine is k8s
                self.remote_engine = self.remote_engine.clone(require_submit=False)
        # we are running at this engine already
        self.host_engine = self.host_engine.clone(require_submit=False)

        # dag_id , execution_date are used by Airflow,
        # should be moved to this class (still used by DB tracking)
        # run.dag_id = AD_HOC_DAG_PREFIX + run.job_name

        self._result_location = None
        self.runtime_errors = []

        # we are running from python notebook, let start to print to stdout
        if run.context.name == "interactive" or is_unit_test_mode():
            run.context.config.set(
                RunLoggingConfig._conf__task_family,
                "stream_stdout",
                "True",
                source="log",
            )

        self._is_initialized = False
        self.task_executor = None

    @property
    def run_settings(self) -> RunSettings:
        return self.databand_context.run_settings

    @property
    def run_config(self) -> RunConfig:
        return self.run_settings.run

    @property
    def env(self) -> EnvConfig:
        return self.run_settings.env

    def _on_enter(self):
        pm.hook.dbnd_on_pre_init_context(ctx=self)
        run_user_func(config.get("core", "user_pre_init"))
        # if we are deserialized - we don't need to run this code again.
        run_config = self.run_config
        if not self._is_initialized:
            # will be called from singleton context manager

            pm.hook.dbnd_on_new_context(ctx=self)

            # RUN USER SETUP FUNCTIONS
            _run_user_func(
                run_config.__class__.user_driver_init, run_config.user_driver_init
            )

            self._is_initialized = True
        else:
            # we get here if we are running at sub process that recreates the Context
            pm.hook.dbnd_on_existing_context(ctx=self)

        # we do it every time we go into databand_config
        self.configure_targets()

        self.run_settings.run_logging.configure_dbnd_logging()

        _run_user_func(run_config.__class__.user_init, run_config.user_init)
        pm.hook.dbnd_post_enter_context(ctx=self)

    def _on_exit(self):
        pm.hook.dbnd_on_exit_context(ctx=self)

    def configure_targets(self):
        output_config = self.run_settings.output
        if output_config.hdf_format == "table":
            register_pd_to_hdf5_as_table_marshaler()

    def is_interactive(self):
        return self.name == "interactive"

    def run_execute(self):
        """
        Runs the main driver!
        """
        run = self.run

        # NO CHANGES TO GLOBAL CONFIG AFTER THIS POINT,
        # run_executor__task will keep the config it was created with, so we will have old config durign .run()
        self.run_executor__task = _RunExecutor_Task(
            task_name=self.run_executor_type, task_version=run.run_uid
        )
        if self.run.root_task:
            # if root_task == None, we will create it in the context of driver task
            # otherwise, we need it to add manually
            self.run_executor__task.descendants.add_child(run.root_task.task_id)

        # will track and "execute" pipeline
        # you need to run DatabandRun.init_run otherwise it will be not tracked
        self.run._driver_task_run = self.run.build_task_run(
            task=self.run_executor__task
        )
        self.build_task_run_executor(
            self.run._driver_task_run, task_engine=self.host_engine
        )

        run.tracker.init_run()
        run.set_run_state(RunState.RUNNING)

        # with captures_log_into_file_as_task_file(log_file=self.local_driver_log.path):
        try:
            if self.run_config.debug_pydevd_pycharm_port is not None:
                debug_port = self.run_config.debug_pydevd_pycharm_port
                # currently no need to export the host to configuration.
                start_pycharm_debugger(host="localhost", port=debug_port)

            self.start_time = utcnow()
            run.driver_task_run.task_run_executor.execute()
            self.finished_time = utcnow()
            # if we are in submitter we don't want to print banner that everything is good
        except DatabandRunError as ex:
            self._dbnd_run_error(ex)
            raise
        except (Exception, KeyboardInterrupt, SystemExit) as ex:
            raise self._dbnd_run_error(ex)
        finally:
            try:
                self.host_engine.cleanup_after_run()
            except Exception:
                logger.exception("Failed to shutdown the current run, continuing")

        return self

    def get_current_dbnd_local_root(self):
        # we should return here the proper engine config, based in which context we run right now
        # it could be submit, driver or task engine
        return self.env.dbnd_local_root

    def _dbnd_run_error(self, ex):
        if (
            # what scenario is this aiflow filtering supposed to help with?
            # I had airflow put a default airflow.cfg in .dbnd causing validation error in k8sExecutor which was invisible in the console (only in task log)
            (
                "airflow" not in ex.__class__.__name__.lower()
                or ex.__class__.__name__ == "AirflowConfigException"
            )
            and "Failed tasks are:" not in str(ex)
            and not isinstance(ex, DatabandRunError)
            and not isinstance(ex, KeyboardInterrupt)
            and not isinstance(ex, DatabandSigTermError)
        ):
            logger.exception(ex)

        if (
            isinstance(ex, KeyboardInterrupt)
            or isinstance(ex, DatabandSigTermError)
            or self.is_killed()
        ):
            run_state = RunState.CANCELLED
            unfinished_task_state = TaskRunState.UPSTREAM_FAILED
        elif isinstance(ex, DatabandFailFastError):
            run_state = RunState.FAILED
            unfinished_task_state = TaskRunState.UPSTREAM_FAILED
        else:
            run_state = RunState.FAILED
            unfinished_task_state = TaskRunState.FAILED

        self.run.set_run_state(run_state)
        self.run.tracker.tracking_store.set_unfinished_tasks_state(
            run_uid=self.run.run_uid, state=unfinished_task_state
        )

        logger.warning(
            "\n\n{sep}\n{banner}\n{sep}".format(
                sep=console_utils.error_separator(),
                banner=self.run.describe.run_banner(
                    "Your run has failed! See more info above.",
                    color="red",
                    show_run_info=True,
                    show_tasks_info=True,
                ),
            )
        )
        return DatabandRunError(
            "Run has failed: %s" % ex, run=self.run, nested_exceptions=ex
        )

    def _is_save_run_pickle(self, task_runs, remote_engine):
        """we need to save run pickle only if we want to execute it in the remote process"""
        if self.run_config.always_save_pipeline:
            return True
        if self.run_config.disable_save_pipeline:
            return False

        if any(tr.task._conf__require_run_dump_file for tr in task_runs):
            return True

        if remote_engine.require_submit:
            return True

        if self.task_executor_type == TaskExecutorType.local:
            return False

        if is_dbnd_orchestration_via_airflow_enabled():
            from dbnd_run.airflow.executors import AirflowTaskExecutorType

            return self.task_executor_type not in [
                AirflowTaskExecutorType.airflow_inprocess,
                TaskExecutorType.local,
            ]
        return True

    def save_run_pickle(self, target_file=None):
        """
        dumps current run and context to file
        """
        t = target_file or self.driver_dump
        logger.info("Saving current run into %s", t)

        # Ensure tracking is completed before pickling
        self.run.tracker.tracking_store.flush()

        with t.open("wb") as fp:
            cloudpickle.dump(obj=self.run.run_executor, file=fp)

    @classmethod
    def load_run(cls, dump_file, disable_tracking_api):
        # type: (FileTarget, bool) -> RunExecutor
        logger.info("Loading dbnd run execution from %s", dump_file)
        with dump_file.open("rb") as fp:
            run_executor: RunExecutor = cloudpickle.load(file=fp)
            if disable_tracking_api:
                run_executor.run_config.context.tracking_store.disable_tracking_api()
                logger.info("Tracking has been disabled")
        try:
            if run_executor.run_config.pickle_handler:
                pickle_handler = load_python_callable(
                    run_executor.run_config.pickle_handler
                )
                pickle_handler(run_executor)
        except Exception as e:
            logger.exception(
                "error while trying to handle pickle with custom handler:", e
            )
        return run_executor

    def is_killed(self):
        return _is_killed.is_set()

    def _internal_kill(self):
        """
        called by TaskRun handler, so we know that run is "canceled"
        otherwise we will get regular exception
        """
        _is_killed.set()

    def kill(self):
        """
        called from user space, kills the current task only
        :return:
        """
        # this is very naive stop implementation
        # in case of simple executor, we'll run task.on_kill code
        _is_killed.set()
        try:
            current_task = None
            from dbnd._core.task_build.task_context import TaskContext, TaskContextPhase

            tc = TaskContext.try_instance()
            if tc.phase == TaskContextPhase.RUN:
                current_list = list(tc.stack)
                if current_list:
                    current_task = current_list.pop()
        except Exception as ex:
            logger.warning("Failed to find current task: %s" % ex)
            return

        if not current_task:
            logger.info("No current task.. Killing nothing..")
            return

        try:
            current_task.on_kill()
        except Exception as ex:
            logger.warning("Failed to kill current task %s: %s" % (current_task, ex))
            return

    def kill_run(self, message=None):
        _is_killed.set()

        # When initiating kill_run, the api's kill_run sends a signal to all running runs,
        # to change their state to shutdown, which in the end sets it to cancelled.
        # the task which initiated the killing, the current task run, should have a state of Failed, and not Canceled.
        # It is important to set it with an error, to allow the passing of error message, to be displayed in the UI
        # as the error_message for the whole run.
        tr = current_task_run()
        if tr.run == self.run:
            task_run_error = TaskRunError.build_from_message(
                task_run=tr,
                msg=message or DEFAULT_TASK_CANCELED_ERR_MSG,
                help_msg="task with task_run_uid:%s initiated kill_run"
                % (tr.task_run_uid),
                ex_class=DbndCanceledRunError,
            )
            tr.set_task_run_state(TaskRunState.FAILED, track=True, error=task_run_error)
        try:
            kill_run(str(self.run.run_uid), ctx=self.run.context)
        except Exception as e:
            raise DatabandFailFastError(
                "Could not send request to kill databand run!", e
            )
        if tr.run == self.run:
            raise DatabandError(message or DEFAULT_TASK_CANCELED_ERR_MSG)

    # all paths, we make them system, we don't want to check if they are exists

    def _init_task_runs_for_execution(self, task_engine):
        """
        creates all relevant task runs starting root_task (driver is not part of this logic)
        we run it from Driver, so this process is logged via _executor task
        """
        run = self.run
        task = run.root_task

        task_runs = TaskRunsBuilder().build_task_runs(
            self, task, task_run_engine=task_engine
        )

        # this one will add tasks to the run! track task runs
        run.tracker.add_task_runs(task_runs)

        return task_runs

    def run_task_at_execution_time(self, task, task_engine=None):
        if task_engine is None:
            task_engine = self.host_engine
        task_run = self.run.build_task_run_and_track(task)
        task_run_executor = self.build_task_run_executor(task_run, task_engine)
        task_run_executor.execute()
        return task_run

    def build_task_run_executor(self, task_run, task_engine):
        task_run_executor = TaskRunExecutor(
            task_run=task_run, run_executor=self, task_engine=task_engine
        )
        task_run.task_run_executor = task_run_executor
        return task_run_executor

    def run_driver(self):
        logger.info("Running driver... Driver PID: %s", os.getpid())

        run = self.run  # type: DatabandRun
        run_executor = run.run_executor
        run_settings = run_executor.run_settings
        remote_engine = run_executor.remote_engine

        run_settings.git.validate_git_policy()
        # let prepare for remote execution
        remote_engine.prepare_for_run(run)

        if self.root_task_name_to_build:

            if self.force_task_name:
                kwargs = {"task_name": self.force_task_name}

                logger.info(
                    "Building main task '%s' with name %s",
                    self.root_task_name_to_build,
                    self.force_task_name,
                )
            else:
                logger.info("Building main task '%s'", self.root_task_name_to_build)
                kwargs = {}
            root_task = get_task_registry().build_dbnd_task(
                self.root_task_name_to_build, task_kwargs=kwargs
            )
            logger.info(
                "Task %s has been created (%s children)",
                root_task.task_id,
                len(root_task.ctrl.task_dag.subdag_tasks()),
            )
            run.root_task = root_task

        tasks = run.root_task.task_dag.subdag_tasks()
        # assert that graph is DAG
        topological_sort(tasks)

        # now we init all task runs for all tasks in the pipeline
        task_runs = self._init_task_runs_for_execution(task_engine=remote_engine)
        root_task_run = run.root_task_run
        run.root_task.ctrl.banner(
            "Main task '%s' has been created!" % root_task_run.task_af_id,
            color="cyan",
            task_run=root_task_run,
        )

        if self.run_config.dry:
            run.root_task.ctrl.describe_dag.describe_dag()
            logger.warning("Execution has been stopped due to run.dry=True flag!")
            return run

        print_tasks_tree(root_task_run.task, task_runs)
        if self._is_save_run_pickle(task_runs, remote_engine):
            run_executor.save_run_pickle()

        task_runs_to_run = [tr for tr in task_runs if not tr.is_skipped]

        # THIS IS THE POINT WHEN WE SUBMIT ALL TASKS TO EXECUTION
        # we should make sure that we create executor without driver task
        self.task_executor = get_task_executor(
            self,
            task_executor_type=run_executor.task_executor_type,
            host_engine=run_executor.host_engine,
            target_engine=remote_engine,
            task_runs=task_runs_to_run,
        )

        hearbeat = None
        if self.send_heartbeat:
            # this will wrap the executor with "heartbeat" process
            hearbeat = start_heartbeat_sender(self)

        with nested(hearbeat):
            self.task_executor.do_run()

        # We need place the pipeline's task_band in the place we required to by outside configuration
        if run_settings.run.run_result_json_path:
            new_path = run_settings.run.run_result_json_path
            try:
                self.result_location.copy(new_path)
            except Exception as e:
                logger.exception(
                    "Couldn't copy the task_band from {old_path} to {new_path}. Failed with this error: {error}".format(
                        old_path=self.result_location.path, new_path=new_path, error=e
                    )
                )

            else:
                logger.info(
                    "Copied the pipeline's task_band to {new_path}".format(
                        new_path=new_path
                    )
                )

        # if we are in the driver, we want to print banner after executor__task banner
        run.set_run_state(RunState.SUCCESS)

        root_task = self.run.root_task_run.task
        msg = "Your run has been successfully executed!"
        if self.run.duration:
            msg = "Your run has been successfully executed in %s" % self.run.duration
        run_msg = "\n%s\n%s\n" % (
            root_task.ctrl.banner(
                "Main task '%s' is ready!" % root_task.task_name,
                color="green",
                task_run=self.run.root_task_run,
            ),
            run.describe.run_banner(msg, color="green", show_tasks_info=True),
        )
        logger.info(run_msg)

        return run

    def run_submitter(self):
        """
        This is the task that represents "submission"
        it can just one task, or.. more tasks, as we can have "docker builds" or other preparations
        this is why we will not run it directly, but do a "full run" with executor

        """
        # we are running submitter, that will send driver to remote
        self.run_settings.git.validate_git_policy()

        run = self.run
        # let prepare for remote execution
        remote_engine = self.remote_engine
        remote_engine.prepare_for_run(run)

        result_map_target = self.run_root.file("{}.json".format(get_uuid()))
        cmd_line_args = (
            ["run"]
            + _get_dbnd_run_relative_cmd()
            + [
                "--run-driver",
                str(run.run_uid),
                "--set",
                "run.run_result_json_path={}".format(result_map_target.path),
                "--set",
                "run_info.execution_date={}".format(
                    run.execution_date.strftime("%Y-%m-%dT%H%M%S.%f")
                ),
            ]
        )

        args = remote_engine.dbnd_executable + cmd_line_args
        submit_to_engine_task = remote_engine.submit_to_engine_task(
            env=run.run_executor.env,
            args=args,
            task_name="dbnd_driver_run",
            interactive=self.run_config.interactive,
        )
        submit_to_engine_task._conf_confirm_on_kill_msg = (
            "Ctrl-C Do you want to kill your submitted pipeline?"
            "If selection is 'no', this process will detach from the run."
        )
        run.root_task = submit_to_engine_task

        # we run all tasks on local engine
        task_runs = self._init_task_runs_for_execution(task_engine=self.host_engine)

        # create executor without driver task!
        # We use local executor to run all tasks (submit_to_engine and required by it tasks)
        # In most cases it will run only submit_to_engine task,
        # But there are scenarios when submit_to_engine task asks for docker builds
        # so we execute the whole pipeline.
        task_executor = LocalTaskExecutor(
            run_executor=self,
            task_executor_type=TaskExecutorType.local,
            host_engine=self.host_engine,
            target_engine=self.host_engine,
            task_runs=task_runs,
        )

        task_executor.do_run()
        self.result_location = result_map_target

        logger.info(run.describe.run_banner_for_submitted())

    def cleanup_after_task_run(self, task):
        # type: (Task) -> None
        rels = task.ctrl.relations
        # potentially, all inputs/outputs targets for current task could be removed
        targets_to_clean = set(flatten([rels.task_inputs, rels.task_outputs]))

        targets_in_use = set()
        # any target which appears in inputs of all not finished tasks shouldn't be removed
        for tr in self.run.task_runs:
            if tr.task_run_state in TaskRunState.final_states():
                continue
            # remove all still needed inputs from targets_to_clean list
            for target in flatten(tr.task.ctrl.relations.task_inputs):
                targets_in_use.add(target)

        TARGET_CACHE.clear_for_targets(targets_to_clean - targets_in_use)

    def get_upstream_failed_task_run_state(self, task_run: "TaskRun") -> "TaskRunState":
        """
        This function is a helper function to decide task final state only for orchestration runs.
        param: task_run_executor - the task  which we want to get it's state
        return: TaskRunState
        """
        state = TaskRunState.UPSTREAM_FAILED
        child_task_runs = task_run.task.descendants
        # Since task run has children, IT'S a pipeline task, so we need to calculate it's state based on children states
        if child_task_runs:
            for task_run_id in child_task_runs.children:
                child_tr_instance = self.run.get_task_run_by_id(task_run_id)
                if child_tr_instance.task_run_state == TaskRunState.FAILED:
                    # In case one of its children fail,task run state is force updated from upstream failed to failed
                    #   because upstream_failed means task dependency is not met, while actually inner task has errors
                    state = TaskRunState.FAILED
        if state == TaskRunState.FAILED:
            logger.info(
                "Task %s set to Failed as one of it's child tasks has been Failed.",
                task_run,
            )

        return state

    def get_context_spawn_env(self):
        env = self.run.get_context_spawn_env()
        if self.run_config.user_code_on_fork:
            env[ENV_DBND__USER_PRE_INIT] = self.run_config.user_code_on_fork
        return env

    @property
    def result(self):
        # type: () -> RunResultBand
        return RunResultBand.from_target(self.result_location)

    @property
    def result_location(self):
        # type: () -> FileTarget
        if self._result_location:
            return target(self._result_location)
        return self.run.root_task.task_band

    @result_location.setter
    def result_location(self, path):
        # type: (FileTarget) -> None
        self._result_location = path

    @property
    def settings(self):
        return self.databand_context.settings

    @property
    def config(self):
        return self.databand_context.config


@contextlib.contextmanager
def set_active_run_context(run):
    # type: (DatabandRun) -> Iterator[DatabandRun]

    from dbnd._core.context.databand_context import DatabandContext  # noqa: F811

    with DatabandContext.context(_context=run.context):
        with DatabandRun.context(_context=run) as dr:
            with RunExecutor.context(_context=run.run_executor) as de:
                yield dr


class _RunExecutor_Task(Task):
    """
    Main purpose of this task is to wrap "RunExecutor" execution
    so we can see logs, some important values as task parameters

    RunExecutor -> _Task -> ... (whatever runs here is available as "DriverTaskRun")

    """

    task_is_system = True
    # we don't want child task (root task) to inherit anything from this one
    _conf__scoped_params = False

    def run(self):
        executor_task_run = current_task_run()
        run_executor = executor_task_run.run.run_executor
        run_executor_type = run_executor.run_executor_type
        try:
            if run_executor_type == SystemTaskName.driver:
                return run_executor.run_driver()
            elif run_executor_type == SystemTaskName.driver_submit:
                return run_executor.run_submitter()
            else:
                raise DatabandSystemError(
                    "Unsupported run executor type: %s" % run_executor_type
                )
        except BaseException as ex:
            # we print it on any exception
            logger.warning("Run failure: %s %s" % (type(ex).__name__, ex))
            logger.warning(
                "\n\n\n\n{sep}\n\n   -= Your run has failed, please review errors below =-\n\n{sep}\n".format(
                    sep=console_utils.error_separator()
                )
            )

            failed_msgs = []
            canceled_msgs = []
            for task_run in executor_task_run.run.get_task_runs():
                if task_run.task_run_state == TaskRunState.FAILED:
                    failed_msgs.append(
                        task_run.task.ctrl.banner(
                            msg="Task has failed!", color="red", task_run=task_run
                        )
                    )
                elif task_run.task_run_state == TaskRunState.CANCELLED:
                    canceled_msgs.append(
                        task_run.task.ctrl.banner(
                            msg="Task has been canceled!",
                            color="yellow",
                            task_run=task_run,
                        )
                    )

            if canceled_msgs:
                logger.warning(
                    "\nNumber of canceled tasks={count}:\n{banner}\n".format(
                        banner="\n".join(canceled_msgs), count=len(canceled_msgs)
                    )
                )

            if failed_msgs:
                logger.warning(
                    "\nNumber of failed tasks={count}:\n{banner}\n".format(
                        banner="\n".join(failed_msgs), count=len(failed_msgs)
                    )
                )
            raise


def dbnd_run_task(
    task_or_task_name: Union[Task, str],
    context,
    job_name: Optional[str] = None,
    force_task_name: Optional[str] = None,
    project: Optional[str] = None,
    run_uid: Optional[UUID] = None,
    existing_run: Optional[bool] = None,
    scheduled_run_info: Optional[ScheduledRunInfo] = None,
    send_heartbeat: bool = True,
) -> RunExecutor:
    """
    This is the main entry point to run task in "dbnd orchestration" mode
    called from `dbnd run`
    we create a new Run + RunExecutor and trigger the execution

    :param task_or_task_name task name to run or already built task object
    :param force_task_name
    :param project Project name for the run
    :return RunExecutor
    """
    job_name = get_task_name_safe(job_name or task_or_task_name)
    project_name = get_project_name_safe(
        project
        or context.config.get("tracking", "project")
        or context.settings.tracking.project,
        task_or_task_name,
    )

    with DatabandRun.new_context(
        context=context,
        job_name=job_name,
        run_uid=run_uid,
        existing_run=existing_run,
        scheduled_run_info=scheduled_run_info,
        is_orchestration=True,
        project_name=project_name,
        allow_override=True,
    ) as run:  # type: DatabandRun
        # this is the main entry point to run some task in "orchestration" mode
        with RunExecutor.new_context(
            run=run,
            root_task_or_task_name=task_or_task_name,
            force_task_name=force_task_name,
            send_heartbeat=send_heartbeat,
            allow_override=True,
        ) as run_executor:
            run.run_executor = run_executor
            run_executor.run_execute()
        return run_executor


def _get_dbnd_run_relative_cmd():
    """returns command without 'dbnd run' prefix"""
    argv = list(sys.argv)
    while argv:
        current = argv.pop(0)
        if current == "run":
            return argv
    raise DatabandRunError(
        "Can't calculate run command from '%s'",
        help_msg="Check that it has a format of '..executable.. run ...'",
    )


def _run_user_func(param, value):
    if not value:
        return
    return run_user_func(value)



# © Copyright Databand.ai, an IBM Company 2022

import contextlib
import logging
import os
import shutil
import subprocess
import sys
import zipfile

from tempfile import mkdtemp

from dbnd._core.current import is_verbose


logger = logging.getLogger(__name__)


def _get_package_name_and_version_from_whl(whl_dir):
    all_files = os.listdir(whl_dir)
    if not all_files:
        raise Exception("No Whl files where created")

    if len(all_files) > 1:
        raise Exception("Directory has more than one whl file")

    whl_file = all_files[0]
    split_name = whl_file.split("-")
    return split_name[0], split_name[1]


def _run_command(generate_command):
    if (
        subprocess.call(generate_command, shell=True) != 0  # nosec B602
    ):  # (Ignore this for bandit as the generated command is not received from the user)
        raise Exception("Failed running {} command".format(generate_command))


@contextlib.contextmanager
def _create_temp_working_dir(tmp_build_dir=None):
    clean_build_dir = False
    try:
        if not tmp_build_dir:
            tmp_build_dir = mkdtemp(prefix="dbnd-build-")
            clean_build_dir = True
        yield tmp_build_dir
    finally:
        if clean_build_dir:
            if is_verbose():  # do not clean build dir in verbose mode
                logger.info("Keeping build dir because verbose mode is on")
            else:
                logger.info("Deleting tmp directory: %s", tmp_build_dir)
                shutil.rmtree(tmp_build_dir, ignore_errors=True)


def generate_project_whl(package_dir, output_dir):
    # generating deps wheels

    # Very important to change to the working dir, otherwise, wheel creation won't work as expected

    curdir = os.curdir
    setup_py_path = os.path.join(package_dir, "setup.py")
    if not os.path.exists(setup_py_path):
        raise Exception(
            "Can't find setup.py inside package dir {} (curdir={})".format(
                package_dir, curdir
            )
        )

    # we are going to be inside the package dir (this way pip works correctly)
    generate_command = "{} {} bdist_wheel --dist-dir {}".format(
        sys.executable, "setup.py", output_dir
    )

    try:
        os.chdir(package_dir)
        _run_command(generate_command)
    finally:
        os.chdir(curdir)


def generate_third_party_deps(requirements_file, output_dir):
    # generating deps wheels
    generate_command = "{} -m pip wheel -r {} -w {}".format(
        sys.executable, requirements_file, output_dir
    )

    _run_command(generate_command)


def zip_dir(zip_file_path, source_dir_path):
    result_zip = zipfile.ZipFile(zip_file_path, "w", zipfile.ZIP_DEFLATED)

    for dir_name, _, files in os.walk(source_dir_path):
        if dir_name == source_dir_path:
            for file_name in files:
                full_file_name = os.path.join(dir_name, file_name)
                temp_zip = zipfile.ZipFile(file=full_file_name, mode="r")
                for file_info in temp_zip.filelist:
                    file_content = temp_zip.read(file_info.filename)
                    result_zip.writestr(file_info, file_content)
    result_zip.close()



# © Copyright Databand.ai, an IBM Company 2022

import logging
import os

from typing import Any, Optional
from uuid import UUID

from airflow.models import BaseOperator

from dbnd import PythonTask, dbnd_bootstrap, new_dbnd_context, override, parameter
from dbnd._core.configuration.environ_config import DBND_RESUBMIT_RUN, DBND_RUN_UID
from dbnd._core.constants import TaskExecutorType
from dbnd._core.run.databand_run import DatabandRun
from dbnd._core.tracking.no_tracking import dont_track
from dbnd._core.tracking.schemas.tracking_info_run import ScheduledRunInfo
from dbnd_run.run_settings import RunConfig
from dbnd_run.run_settings.log import RunLoggingConfig
from dbnd_run.tasks.basics.shell import bash_cmd


logger = logging.getLogger(__name__)


@dont_track
class DbndSchedulerOperator(BaseOperator):
    template_fields = ("scheduled_cmd",)
    template_ext = (".sh", ".bash")
    ui_color = "#f0ede4"

    def __init__(
        self,
        scheduled_cmd,
        scheduled_job_name,
        scheduled_job_uid,
        shell,
        with_name=True,
        extra_args=None,
        **kwargs
    ):
        # type: (str, str, Optional[UUID], bool, bool, str, **Any) ->  DbndSchedulerOperator
        super(DbndSchedulerOperator, self).__init__(**kwargs)
        self.scheduled_job_name = scheduled_job_name
        self.scheduled_job_uid = scheduled_job_uid
        self.scheduled_cmd = scheduled_cmd
        self.shell = shell
        self.with_name = with_name
        self.extra_args = extra_args

    def execute(self, context):
        scheduled_run_info = ScheduledRunInfo(
            scheduled_job_uid=self.scheduled_job_uid,
            scheduled_job_dag_run_id=context.get("dag_run").id,
            scheduled_date=context.get("task_instance").execution_date,
            scheduled_job_name=self.scheduled_job_name if self.with_name else None,
            scheduled_job_extra_args=self.extra_args,
        )
        # disable heartbeat at this level,
        # otherwise scheduled jobs that will run on Kubernetes
        # will always have a heartbeat even if the actual driver
        # sent to Kubernetes is lost down the line,
        # which is the main purpose of the heartbeat

        self.launch(context, scheduled_run_info)

    def launch(self, context, scheduled_run_info: ScheduledRunInfo):
        dbnd_bootstrap(enable_dbnd_run=True)
        with new_dbnd_context(
            name="airflow",
            conf={
                RunConfig.task_executor_type: override(TaskExecutorType.local),
                RunConfig.parallel: override(False),
                RunLoggingConfig.disabled: override(True),
            },
        ) as dc:
            launcher_task = Launcher(
                scheduled_cmd=self.scheduled_cmd,
                task_name=context.get("dag").dag_id,
                task_version="now",
                task_is_system=True,
                shell=self.shell,
            )

            dc.dbnd_run_task(
                task_or_task_name=launcher_task,
                scheduled_run_info=scheduled_run_info,
                send_heartbeat=False,
            )


class Launcher(PythonTask):
    scheduled_cmd = parameter[str]
    shell = parameter[bool]

    def run(self):
        env = os.environ.copy()
        env[DBND_RUN_UID] = str(DatabandRun.get_instance().run_uid)
        env[DBND_RESUBMIT_RUN] = "true"

        return bash_cmd.callable(
            cmd=self.scheduled_cmd, env=env, dbnd_env=False, shell=self.shell
        )



# © Copyright Databand.ai, an IBM Company 2022

import logging
import os

from typing import List

from dbnd import output, parameter, task
from targets import Target


logger = logging.getLogger(__name__)


@task(result=List[str])
def publish_results(data=parameter[Target], published=output.folder[Target]):
    if isinstance(data, dict):
        data = data.items()
    for partition in data:
        if isinstance(partition, tuple):
            partition_basename, partition = partition
            partition_basename = str(partition_basename)
            original_ext = os.path.splitext(os.path.basename(partition.path))[1]
            if not os.path.splitext(partition_basename)[1]:
                logger.info("Partition %s doesn't have extension")
                partition_basename += original_ext
        else:
            partition_basename = os.path.basename(partition.path)
        target_path = os.path.join(published.path, partition_basename)
        logger.info("Copying %s to %s", partition, target_path)
        partition.copy(target_path)
    published.mark_success()
    return [str(partition) for partition in data]



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd._core.constants import TaskRunState
from dbnd._core.errors.base import DatabandRunError, DatabandSigTermError
from dbnd_run import errors
from dbnd_run.run_executor_engine import RunExecutorEngine


logger = logging.getLogger(__name__)


class LocalTaskExecutor(RunExecutorEngine):
    def do_run(self):
        topological_tasks = topological_sort([tr.task for tr in self.task_runs])
        fail_fast = self.run_executor.run_config.fail_fast
        task_failed = False

        task_runs_to_update_state = []
        for task in topological_tasks:
            tr = self.run.get_task_run_by_id(task.task_id)
            if tr.is_reused:
                continue

            if fail_fast and task_failed:
                state = self.run_executor.get_upstream_failed_task_run_state(tr)

                logger.info("Setting %s to %s", task.task_id, state)
                tr.set_task_run_state(state, track=False)
                task_runs_to_update_state.append(tr)
                continue

            upstream_task_runs = [
                self.run.get_task_run_by_id(t.task_id)
                for t in task.ctrl.task_dag.upstream
            ]
            failed_upstream = [
                upstream_task_run
                for upstream_task_run in upstream_task_runs
                if upstream_task_run.task_run_state in TaskRunState.fail_states()
            ]
            if failed_upstream:
                logger.info(
                    "Setting %s to %s", task.task_id, TaskRunState.UPSTREAM_FAILED
                )
                tr.set_task_run_state(TaskRunState.UPSTREAM_FAILED, track=False)
                task_runs_to_update_state.append(tr)
                continue

            if self.run_executor.is_killed():
                logger.info(
                    "Databand Context is killed! Stopping %s to %s",
                    task.task_id,
                    TaskRunState.FAILED,
                )
                tr.set_task_run_state(TaskRunState.FAILED, track=False)
                task_runs_to_update_state.append(tr)
                continue

            logger.debug("Executing task: %s", task.task_id)

            try:
                tr.task_run_executor.execute()
            except DatabandSigTermError as e:
                raise e
            except Exception:
                task_failed = True
                logger.exception("Failed to execute task '%s':" % task.task_id)

        if task_runs_to_update_state:
            self.run.tracker.set_task_run_states(task_runs_to_update_state)

        if task_failed:
            err = _collect_errors(self.run.task_runs)

            if err:
                raise DatabandRunError(err)


def _collect_errors(task_runs):
    err = ""
    upstream_failed = []
    failed = []
    for task_run in task_runs:
        task_name = task_run.task.task_name
        if task_run.task_run_state == TaskRunState.UPSTREAM_FAILED:
            # we don't want to show upstream failed in the list
            upstream_failed.append(task_name)
        elif task_run.task_run_state in TaskRunState.direct_fail_states():
            failed.append(task_name)
    if upstream_failed:
        err += "Task that didn't run because of failed dependency:\n\t{}\n".format(
            "\n\t".join(upstream_failed)
        )
    if failed:
        err += "Failed tasks are:\n\t{}".format("\n\t".join(failed))
    return err


def topological_sort(tasks, root_task=None):
    """
    Sorts tasks in topographical order, such that a task comes after any of its
    upstream dependencies.

    Heavily inspired by:
    http://blog.jupo.org/2012/04/06/topological-sorting-acyclic-directed-graphs/

    :return: list of tasks in topological order
    """

    # special case
    if len(tasks) == 0:
        return tuple()

    graph_unsorted = set(tasks)

    graph_sorted = []

    # Run until the unsorted graph is empty.
    while graph_unsorted:
        # Go through each of the node/edges pairs in the unsorted
        # graph. If a set of edges doesn't contain any nodes that
        # haven't been resolved, that is, that are still in the
        # unsorted graph, remove the pair from the unsorted graph,
        # and append it to the sorted graph. Note here that by using
        # using the items() method for iterating, a copy of the
        # unsorted graph is used, allowing us to modify the unsorted
        # graph as we move through it. We also keep a flag for
        # checking that that graph is acyclic, which is true if any
        # nodes are resolved during each pass through the graph. If
        # not, we need to bail out as the graph therefore can't be
        # sorted.
        acyclic = False
        for node in list(graph_unsorted):
            for edge in node.ctrl.task_dag.upstream:
                if edge in graph_unsorted:
                    break
            # no edges in upstream tasks
            else:
                # we found at least one node in graph
                acyclic = True
                graph_unsorted.remove(node)
                graph_sorted.append(node)

        if not acyclic:
            raise errors.graph.cyclic_graph_detected(root_task, graph_unsorted)

    return tuple(graph_sorted)



# © Copyright Databand.ai, an IBM Company 2022

# we need to import databand module before airflow, otherwise we will not get airflow_bome

import logging
import os

import pytest

from dbnd import config, dbnd_bootstrap, dbnd_config
from dbnd._core.configuration.environ_config import reset_dbnd_project_config
from dbnd._core.context.databand_context import DatabandContext, new_dbnd_context
from dbnd._core.context.use_dbnd_run import set_orchestration_mode
from dbnd._core.task_build.task_registry import tmp_dbnd_registry
from dbnd._core.task_build.task_signature import TASK_ID_INVALID_CHAR_REGEX


logger = logging.getLogger(__name__)


def _run_name_for_test_request(request):
    return TASK_ID_INVALID_CHAR_REGEX.sub("_", request.node.name)


@pytest.fixture
def dbnd_clean_project_config():
    reset_dbnd_project_config()

    # dbnd_bootstrap is "sinlge execution",
    # it will run only once per pytest proces.

    dbnd_bootstrap(enable_dbnd_run=True)

    set_orchestration_mode()
    dbnd_config.reset()
    dbnd_config.load_system_configs()

    with tmp_dbnd_registry() as r:
        yield r

    reset_dbnd_project_config()


@pytest.fixture
def dbnd_clean_airflow_vars_for_every_test():
    yield
    for v in [
        "AIRFLOW_CTX_DAG_ID",
        "AIRFLOW_CTX_EXECUTION_DATE",
        "AIRFLOW_CTX_TASK_ID",
        "AIRFLOW_CTX_TRY_NUMBER",
        "AIRFLOW_CTX_UID",
    ]:
        if v in os.environ:
            del os.environ[v]


@pytest.fixture
def dbnd_config_for_test_run__system(request, tmpdir):
    return {
        "run_info": {"name": _run_name_for_test_request(request)},
        "run": {"heartbeat_interval_s": -1},
        "local": {"root": str(tmpdir.join("local_root"))},
    }


@pytest.fixture
def dbnd_context_kwargs():
    return {}


@pytest.fixture
def dbnd_config_for_test_run__user():
    # USE IT TO OVEERRIDE/ADD EXTRA CONFIG
    return {}


@pytest.fixture
def dbnd_run_test_context(
    dbnd_clean_project_config,
    dbnd_context_kwargs,
    dbnd_config_for_test_run__system,
    dbnd_config_for_test_run__user,
):  # type: (...) -> DatabandContext

    dbnd_context_kwargs.setdefault(
        "name", f"pytest fixture context: {os.environ.get('PYTEST_CURRENT_TEST')}"
    )

    with config(
        dbnd_config_for_test_run__system, source="dbnd_config_for_test_run__system"
    ), config(
        dbnd_config_for_test_run__user, source="dbnd_config_for_test_run__user"
    ), new_dbnd_context(
        **dbnd_context_kwargs
    ) as t:
        yield t


@pytest.fixture
def dbnd_run_pytest_env(dbnd_clean_airflow_vars_for_every_test, dbnd_run_test_context):
    """
    all required fixtures to run datband related tests
    add it to your [pytest]usefixtures
    """
    return True



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.parameter.parameter_builder import parameter
from dbnd._core.task import Config


class DescribeConfig(Config):
    """(Advanced) Databand's --describe behaviour"""

    _conf__task_family = "describe"

    dry = parameter(
        default=False, description="Describe without pushing to databand-web"
    )[bool]

    no_checks = parameter(
        default=False, description="Describe without doing copleteness and other checks"
    )[bool]
    no_tree = parameter(
        default=False, description="Describe without showing tasks tree"
    )[bool]

    print_task_band = parameter(description="Print task_band in logging output.").value(
        True
    )



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.errors import DatabandError


def failed_to_load_versioned_dagbag_plugin(exc):
    return DatabandError(
        "Failed to switch to versioned dagbag!", exc, show_exc_info=True
    )


def failed_to_retrieve_dag_via_dbnd_versioned_dagbag(exc):
    return DatabandError(
        "Failed to retrieve DAG from DbndDagBag!", exc=exc, show_exc_info=True
    )



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.task_build.task_const import _SAME_AS_PYTHON_MODULE
from dbnd._core.task_build.task_registry import get_task_registry


def namespace(namespace=None, scope=""):
    """
    Call to set namespace of tasks declared after the call.

    It is often desired to call this function with the keyword argument
    ``scope=__name__``.

    The ``scope`` keyword makes it so that this call is only effective for task
    classes with a matching ``__module__``. The default value for
    ``scope`` is the empty string, which means all classes. Multiple calls with
    the same scope simply replace each other.

    The namespace of a ``Task`` can also be changed by specifying the property
    ``task_namespace``.

    ::

        class Task2(dbnd.Task):
            task_namespace = 'namespace2'

    This explicit setting takes priority over whatever is set in the
    ``namespace()`` method, and it's also inherited through normal python
    inheritence.

    There's no equivalent way to set the ``task_family``.
    """
    get_task_registry().register_namespace(scope=scope, namespace=namespace or "")


def auto_namespace(scope=""):
    """
    Same as ``namespace``, but it will be set to the ``__module__`` of the task class.

    This is desirable for these reasons:

    *   Two tasks with the same name will not have conflicting task families
    *   It's more pythonic, as modules are Python's recommended way to do namespacing.
    *   It's traceable. When you see the full name of a task, you can
        immediately identify where it is defined.

    We recommend calling this function from your package's outermost
    ``__init__.py`` file. The file contents could look like this::

        import dbnd

        databand.auto_namespace(scope=__name__)

    To reset an ``auto_namespace()`` call, you can use ``namespace(scope='my_scope')``.
    But this will not be needed (and is also discouraged) if you use the ``scope`` kwarg.
    """
    namespace(namespace=_SAME_AS_PYTHON_MODULE, scope=scope)



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd._core.constants import ClusterPolicy
from dbnd_run.task_ctrl import _TaskRunExecutorCtrl


logger = logging.getLogger(__name__)


class TaskSubEnginePolicyCtrl(_TaskRunExecutorCtrl):
    def __init__(self, task, job):
        super(TaskSubEnginePolicyCtrl, self).__init__(task_run=job)
        self.job = job

    @classmethod
    def create_engine(cls):
        return None

    @classmethod
    def terminate_engine(cls):
        return None

    @classmethod
    def get_engine_policy(cls):
        return ClusterPolicy.NONE

    @classmethod
    def apply_engine_policy(cls, root_task):
        policy = cls.get_engine_policy()
        if policy == ClusterPolicy.NONE:
            return root_task

        # now we support only google cloud engine
        if policy == ClusterPolicy.NONE:
            return root_task

        if policy in [ClusterPolicy.CREATE, ClusterPolicy.EPHERMAL]:
            create_task = cls.create_engine()
            if create_task:
                root_task.set_global_upstream(create_task)

        if policy in [ClusterPolicy.KILL, ClusterPolicy.EPHERMAL]:
            delete_cluster = cls.terminate_engine()
            if delete_cluster:
                root_task.set_downstream(delete_cluster)

        return root_task



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.parameter.parameter_builder import parameter
from dbnd._core.task import config


class SchedulerConfig(config.Config):
    """(Advanced) Databand's scheduler"""

    _conf__task_family = "scheduler"

    config_file = parameter(
        default=None,
        description="Set a path to the file defining scheduled jobs to execute.",
    )[str]

    # by default the scheduler drop-in will decide whether to sync or not based on if it's running inside the scheduler or not (and not the websrver)
    # the next two params can be used to force it one way or the other
    never_file_sync = parameter(
        default=False,
        description="Disable syncing the scheduler config_file to the database.",
    )[bool]

    always_file_sync = parameter(
        default=False,
        description="Enable forcibly syncing the scheduler config_file to the database.",
    )[bool]

    no_ui_cli_edit = parameter(
        default=False,
        description="Disable creating, editing, and deleting scheduled jobs from the CLI and UI. Scheduled job "
        "definitions will only be taken from the scheduler config file.",
    )

    refresh_interval = parameter(
        default=1,
        description="Set the interval to refresh the scheduled job list from the db and/or a config file",
    )[int]

    active_by_default = parameter(
        default=True,
        description="Determine whether new scheduled jobs will be activated by default.",
    )[bool]

    default_retries = parameter(
        description="Set the number of times to retry a failed run, "
        "unless set to a different value on the scheduled job"
    )[int]

    shell_cmd = parameter(
        description="If shell_cmd is True, the specified command will be executed through the shell. "
        "This can be useful if you are using Python primarily "
        "for the enhanced control flow it offers "
        "over most system shells and still want convenient access to other shell features "
        "such as shell pipes, filename wildcards, environment variable expansion, "
        "and expansion of ~ to a user's home directory."
    )[bool]



# © Copyright Databand.ai, an IBM Company 2022

import logging
import os.path
import subprocess
import tarfile
import tempfile

from dbnd import output, task
from dbnd._core.constants import CloudType
from dbnd._core.errors import DatabandRuntimeError
from dbnd._core.utils.timezone import utcnow
from targets.types import Path


logger = logging.getLogger(__name__)


@task(archive=output(output_ext=".tar.gz")[Path])
def export_db(
    archive,
    include_db=True,
    include_logs=True,
    task_version=utcnow().strftime("%Y%m%d_%H%M%S"),
):
    # type: (Path, bool, bool, str)-> None

    from dbnd._core.current import get_databand_context

    logger.info("Compressing files to %s..." % archive)
    with tarfile.open(str(archive), "w:gz") as tar:

        if include_db:
            dbnd_context = get_databand_context()
            conn_string = dbnd_context.config.get("webserver", "sql_alchemy_conn")
            if conn_string.startswith("postgresql"):
                with tempfile.NamedTemporaryFile(prefix="dbdump.", suffix=".sql") as tf:
                    dump_postgres(conn_string, tf.name)
                    tar.add(tf.name, arcname="postgres-dbnd.sql")
            else:
                raise DatabandRuntimeError(
                    "Can not export db! "
                    "Currently, we support only postgres db in automatic export"
                )

        if include_logs:
            context = get_databand_context()
            local_env = context.settings.get_env_config(CloudType.local)
            logs_folder = local_env.dbnd_local_root.folder("logs").path
            if os.path.exists(logs_folder):
                logger.info("Adding run folder from '%s'", logs_folder)
                tar.add(logs_folder, "run")
            else:
                logger.warning("Logs dir '%s' is not found", logs_folder)


def dump_postgres(conn_string, dump_file):
    logger.info(
        "backing up postgres DB to %s, pg_dump and sqlalchemy are required!", dump_file
    )

    from sqlalchemy.engine.url import make_url

    url = make_url(conn_string)

    cmd = [
        "pg_dump",
        "-h",
        url.host,
        "-p",
        str(url.port),
        "-U",
        url.username,
        "-Fc",
        "-f",
        dump_file,
        "-d",
        url.database,
    ]
    logger.info("Running command: %s", subprocess.list2cmdline(cmd))
    env = os.environ.copy()
    env["PGPASSWORD"] = url.password
    subprocess.check_call(args=cmd, env=env)



# © Copyright Databand.ai, an IBM Company 2022
from dbnd_run.cli.cmd_execute import execute
from dbnd_run.cli.cmd_heartbeat import send_heartbeat
from dbnd_run.cli.cmd_project import project_init
from dbnd_run.cli.cmd_run import cmd_run
from dbnd_run.cli.cmd_scheduler_management import schedule
from dbnd_run.cli.cmd_show import show_tasks
from dbnd_run.cli.cmd_utils import collect_logs


def add_dbnd_run_cli(cli):
    # project
    cli.add_command(project_init)

    # run
    cli.add_command(cmd_run)
    cli.add_command(execute)

    # show
    cli.add_command(show_tasks)

    # heartbeat sender
    cli.add_command(send_heartbeat)

    cli.add_command(schedule)

    # error reporting
    cli.add_command(collect_logs)



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.utils.basics.path_utils import relative_path


def get_dbnd_run_conf_file(*path):
    return relative_path(__file__, "..", "conf", *path)



# © Copyright Databand.ai, an IBM Company 2022
from dbnd._core.context.bootstrap import dbnd_bootstrap
from dbnd._core.current import get_databand_context
from dbnd._core.utils.basics import nested_context
from dbnd._vendor import click
from dbnd._vendor.click import command


@command()
@click.option("--run-uid", required=True)
@click.option("--databand-url", required=False)
@click.option("--heartbeat-interval", required=True, type=int)
@click.option("--driver-pid", required=True, type=int)
@click.option("--tracker", required=True)
@click.option("--tracker-api", required=True)
def send_heartbeat(
    run_uid, databand_url, heartbeat_interval, driver_pid, tracker, tracker_api
):

    dbnd_bootstrap(enable_dbnd_run=True)

    from dbnd import config
    from dbnd_run.run_executor.heartbeat_sender import send_heartbeat_continuously

    with config(
        {
            "core": {
                "tracker": tracker.split(","),
                "tracker_api": tracker_api,
                "databand_url": databand_url,
            }
        }
    ):
        requred_context = []
        if tracker_api == "db":
            from dbnd import new_dbnd_context

            requred_context.append(new_dbnd_context(name="send_heartbeat"))

        with nested_context.nested(*requred_context):
            tracking_store = get_databand_context().tracking_store

            send_heartbeat_continuously(
                run_uid, tracking_store, heartbeat_interval, driver_pid
            )



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd._core.errors import DatabandBuildError
from dbnd._core.errors.friendly_error.helpers import _band_call_str


logger = logging.getLogger()


def _find_cirlce(cyclic_nodes):
    # we are trying to find "real" circle our of this graph
    # so we traverse the graph till we find the "back" loop
    # and print only loop part of the graph
    #  A -> B -> C-> D ->..(B)
    #  we will print only  B->C->D..
    graph_unsorted = list(cyclic_nodes)
    # randomly select first node
    node = graph_unsorted[0]
    circle_trail = []

    while graph_unsorted:
        graph_unsorted.remove(node)
        circle_trail.append(node)
        for edge in node.ctrl.task_dag.upstream:
            if edge in circle_trail:
                # we found the circle, let return only relevant nodes
                return circle_trail[circle_trail.index(edge) :]

            # let just keep looking for the trail by continuing into the graph
            if edge in graph_unsorted:
                next_node = edge
                break
        else:
            next_node = graph_unsorted[0]
        node = next_node

    # we should not get here..
    return circle_trail


def cyclic_graph_detected(task, cyclic_nodes):
    # print the line of tasks
    circle_trail = _find_cirlce(cyclic_nodes)

    from dbnd_run.task_ctrl.task_dag_describe import tasks_trail

    logger.warning("Cyclic graph detected: %s", tasks_trail(circle_trail))
    return DatabandBuildError(
        "A cyclic dependency occurred %s: {short_trail}.. ({num_of_tasks} tasks)".format(
            task_call="in '{task_call}'" % _band_call_str(task) if task else "",
            short_trail=tasks_trail(circle_trail[:3]),
            num_of_tasks=len(circle_trail),
        ),
        show_exc_info=False,
        help_msg="Check your %s logic, see the full circle path in log. "
        "For better visability, use run.recheck_circle_dependencies=True"
        % _band_call_str(task),
    )



# © Copyright Databand.ai, an IBM Company 2022

import logging

from warnings import warn

from dbnd._core.configuration.environ_config import ENV_DBND__TRACKING
from dbnd._core.context.bootstrap import dbnd_bootstrap
from dbnd._core.run.databand_run import DatabandRun
from dbnd._core.task_build.task_context import TaskContextPhase
from dbnd._core.utils.basics.environ_utils import env as env_context
from dbnd._vendor import click
from dbnd_run.orchestration_bootstrap import dbnd_disable_databand_dags_loading
from dbnd_run.run_executor.run_executor import RunExecutor, set_active_run_context


logger = logging.getLogger(__name__)


def get_dbnd_version():
    import dbnd

    return dbnd.__version__


def get_python_version():
    import sys

    return sys.version.rsplit(" ")[0]  # string like '3.8.6'


class DbndVersionsClashWarning(UserWarning):
    pass


@click.group()
@click.option("--dbnd-run", required=True, type=click.Path())
# option that disables the tracking store access.
@click.option("--disable-tracking-api", is_flag=True, default=False)
@click.option("--expected-dbnd-version", default=None)
@click.option("--expected-python-version", default=None)
@click.pass_context
def execute(
    ctx, dbnd_run, disable_tracking_api, expected_dbnd_version, expected_python_version
):
    """Execute databand primitives"""
    dbnd_bootstrap(enable_dbnd_run=True)

    from dbnd import config

    dbnd_disable_databand_dags_loading(dbnd_config=config)

    if expected_python_version and expected_dbnd_version:
        spark_python_version = get_python_version()
        spark_dbnd_version = get_dbnd_version()
        if expected_python_version != spark_python_version:
            warn(
                "You submitted job using Python {} but the Spark cluster uses Python {}. To "
                "assure execution consistency use the same version in both places. Execution will"
                "continue but it may fail due to version mismatch.".format(
                    expected_python_version, spark_python_version
                ),
                DbndVersionsClashWarning,
            )
        if expected_dbnd_version != spark_dbnd_version:
            warn(
                "You submitted job using dbnd {} but the Spark cluster uses dbnd {}. To "
                "assure execution consistency use the same version in both places. Execution will"
                "continue but it may fail due to version mismatch.".format(
                    expected_dbnd_version, spark_dbnd_version
                ),
                DbndVersionsClashWarning,
            )

    from targets import target

    with env_context(**{ENV_DBND__TRACKING: "False"}):
        run_executor = RunExecutor.load_run(
            dump_file=target(dbnd_run), disable_tracking_api=disable_tracking_api
        )
        ctx.obj = {
            "run_executor": run_executor,
            "run": run_executor.run,
            "disable_tracking_api": disable_tracking_api,
        }


@execute.command(name="task")
@click.option("--task-id", required=True)
@click.pass_context
def run_task(ctx, task_id):
    """(Internal) Run a task inline (task.run function)"""

    run = ctx.obj["run"]  # type: DatabandRun
    with set_active_run_context(run):
        task = run._get_task_by_id(task_id)
        task_run = task.current_task_run
        # this tracking store should be the same object as the one in the context but they're actually
        # different.
        if ctx.obj["disable_tracking_api"]:
            task_run.tracker.tracking_store.disable_tracking_api()
        with task.ctrl.task_context(phase=TaskContextPhase.RUN):
            task._task_run()


@execute.command(name="task_submit")
@click.option("--task-id", required=True)
@click.pass_context
def run_task_submit(ctx, task_id):
    """Submit a task"""

    run = ctx.obj["run"]  # type: DatabandRun
    with set_active_run_context(run):
        task = run._get_task_by_id(task_id)
        task._task_submit()


@execute.command(name="task_execute")
@click.option("--task-id", required=True)
@click.pass_context
def run_task_execute(ctx, task_id):
    """Execute a task"""

    run = ctx.obj["run"]  # type: DatabandRun
    with set_active_run_context(run):
        task_run = run.get_task_run_by_id(task_id)
        task_run.task_run_executor.execute(allow_resubmit=False)



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd._core.errors import DatabandBuildError
from dbnd._core.parameter.parameter_builder import parameter
from dbnd._core.task import config
from dbnd._core.utils.git import is_git_dirty


logger = logging.getLogger(__name__)


class GitConfig(config.Config):
    """Databand's git validator"""

    _conf__task_family = "git"

    enforce_clean = parameter(
        description="Enforce the project's git to be clean. Can be overridden by `allow_dirty` or `--git-allow-dirty`"
    ).value(False)

    allow_dirty = parameter(
        description="Permit git to be dirty when `enforce_clean` or `--git-enforce-clean` is on"
    ).value(False)

    def _raise_enforce_clean_error(self, msg):
        help_text = (
            "\nYou see this message because enforce_clean in git section is enabled."
            "\nTo temporarily disable this message use --git-allow-dirty."
        )
        raise DatabandBuildError(help_msg=msg + help_text)

    def validate_git_policy(self):
        if not self.enforce_clean:
            return

        is_dirty = is_git_dirty()

        if is_dirty is False:
            return

        if is_dirty:
            if self.allow_dirty:
                logger.warning("Runing with not commited files.")
            else:
                self._raise_enforce_clean_error("Git repo must be clean.")
        else:  # is_dirty is None
            if self.allow_dirty:
                logger.warning("Failed to get git status.")
            else:
                self._raise_enforce_clean_error("Failed to get git status.")



import os


def _default_getuser():
    return "unknown"


def find_runnable_getuser_function():
    try:

        if hasattr(os, "getlogin"):
            os.getlogin()
            return os.getlogin  # check before we return it

        import getpass

        if not hasattr(getpass, "getuser"):
            return _default_getuser()

        getpass.getuser()  # check before we return it
        return getpass.getuser
    except Exception:
        return _default_getuser



# © Copyright Databand.ai, an IBM Company 2022

import logging

from airflow.plugins_manager import AirflowPlugin

from dbnd_run.airflow.web.airflow_app_with_versioned_dagbag import (
    patch_airflow_create_app,
)


# WE SHOULD NOT HAVE ANY dbnd imports here -- circle import:  dbnd_airflow -> airflow -> load .. -> import dbnd_airflow
logger = logging.getLogger()


class DatabandAirflowWebserverPlugin(AirflowPlugin):
    name = "databand_airflow_plugin"

    @classmethod
    def on_load(cls, *args, **kwargs):
        patch_airflow_create_app()



# © Copyright Databand.ai, an IBM Company 2022

import logging

from airflow.models import DAG, BaseOperator

from dbnd import PipelineTask, Task, output, parameter


logger = logging.getLogger(__name__)


def bind_task_to_op(task, airflow_operator):
    task.ctrl.airflow_op = airflow_operator


class AirflowOperatorAsDbndTask(Task):
    """
    Every Airflow Native Operator will be wrapped with AirflowTask
    """

    airflow_task_id = parameter[str]

    task_target_date = None

    airflow_op = parameter(significant=False)[BaseOperator]
    dag = parameter(significant=False)[DAG]

    original_execute = None
    _completed_airflow = False

    def _complete(self):
        return self._completed_airflow

    @classmethod
    def build_airflow_task(cls, airflow_operator, dag):
        task = cls(
            task_name=airflow_operator.task_id,
            airflow_task_id=airflow_operator.task_id,
            airflow_op=airflow_operator,
            dag=dag,
        )
        task.ctrl.airflow_op = airflow_operator
        airflow_operator.dbnd_task_id = task.task_id
        return task

    def run(self):
        af_task = self.airflow_op

        # TODO: use mechanism for "system" metrics
        for attr in af_task.__class__.template_fields:
            content = getattr(af_task, attr)
            if content and attr != "env":
                self.log_metric("airflow__%s" % attr, content)

        airflow_context = self.current_task_run.airflow_context
        assert airflow_context
        result = self.airflow_op.execute(airflow_context)
        self._completed_airflow = True
        return result


class AirflowDagAsDbndTask(PipelineTask):
    dag_id = parameter[str]

    # `get_runnable_airflow_dag` check if the task is AirflowDagTask
    # if it's AirflowDagTask  - we will use original DAG
    dag = parameter(significant=False)[DAG]
    roots = output

    def band(self):
        tasks = {}
        for airflow_op in self.dag.tasks:
            task = AirflowOperatorAsDbndTask.build_airflow_task(
                airflow_operator=airflow_op, dag=self.dag
            )
            tasks[airflow_op.task_id] = task

        for task in tasks.values():
            task.set_upstream(
                [tasks[op.task_id] for op in task.airflow_op.upstream_list]
            )

        self.roots = [tasks[r.task_id] for r in self.dag.roots]

    @classmethod
    def build_dbnd_task_from_dag(cls, dag):
        return cls(task_name=dag.dag_id, dag=dag, dag_id=dag.dag_id)



# © Copyright Databand.ai, an IBM Company 2022

from typing import Tuple


def partition_from_module_name(
    module_name: str, base_filename: str = "dbnd_dags_from_databand_"
) -> Tuple[int, int]:
    index = module_name.find(base_filename)
    parts = module_name[index + len(base_filename) :]
    result = parts.split("__")
    return int(result[0]), int(result[1])



# © Copyright Databand.ai, an IBM Company 2022

import logging
import sys
import typing

from typing import Union

from dbnd._core.parameter.parameter_builder import parameter
from dbnd._core.task import config
from dbnd._core.task_build.task_registry import build_task_from_config
from dbnd_run.errors.executor_k8s import local_engine_not_accept_remote_jobs
from targets import DirTarget


if typing.TYPE_CHECKING:
    from dbnd._core.run.databand_run import DatabandRun

logger = logging.getLogger(__name__)


class EngineConfig(config.Config):
    """Databand's engine configuration (where tasks are executed)"""

    require_submit = parameter(
        description="Should the task engine be forced to submit tasks"
    ).value(False)

    dbnd_local_root = parameter(
        default=None, description="Local dbnd home directory at the engine environment"
    )[DirTarget]

    dbnd_executable = parameter(
        default=[sys.executable, "-m", "dbnd"],
        description="'dbnd' executable path at engine environment",
    )[typing.List[str]]

    def cleanup_after_run(self):
        pass

    def submit_to_engine_task(self, env, task_name, args, interactive=True):
        raise local_engine_not_accept_remote_jobs(self.env, self)

    def prepare_for_run(self, run):
        # type: (DatabandRun) -> None
        return

    def _should_wrap_with_submit_task(self, task_run):
        return self.require_submit


class LocalMachineEngineConfig(EngineConfig):
    """
    Engine configuration for executing on the local machine.
    """

    _conf__task_family = "local_machine"

    def submit_to_engine_task(self, env, task_name, args, interactive=True):
        from dbnd_run.tasks.basics.shell import bash_cmd

        return bash_cmd.task(
            args=args,
            task_version="now",
            task_env=env,
            task_name=task_name,
            task_is_system=True,
        )

    def _should_wrap_with_submit_task(self, task_run):
        """
        We don't want to resubmit if it's bash cmd already
        """
        from dbnd_run.tasks.basics.shell import bash_cmd

        if isinstance(task_run.task, bash_cmd.task):
            return False
        return super(LocalMachineEngineConfig, self)._should_wrap_with_submit_task(
            task_run
        )


def build_engine_config(name):
    # type: ( Union[str, EngineConfig]) -> EngineConfig
    """
    Builds EngineConfig object for `name`
    """
    return build_task_from_config(name, EngineConfig)
